# MicroPython lora reliable_delivery example - synchronous receiver program
# MIT license; Copyright (c) 2023 Angus Gratton
import struct
import time
import machine
from machine import SPI, Pin
from micropython import const
from lora import RxPacket

from lora_rd_settings import RECEIVER_ID, ACK_LENGTH, ACK_DELAY_MS, lora_cfg

# Change _DEBUG to const(True) to get some additional debugging output
# about timing, RSSI, etc.
#
# For a lot more debugging detail, go to the modem driver and set _DEBUG there to const(True)
_DEBUG = const(False)

# Keep track of the last counter value we got from each known sender
# this allows us to tell if packets are being lost
last_counters = {}


def get_modem():
    # from lora import SX1276
    # return SX1276(
    #     spi=SPI(1, baudrate=2000_000, polarity=0, phase=0,
    #             miso=Pin(19), mosi=Pin(27), sck=Pin(5)),
    #     cs=Pin(18),
    #     dio0=Pin(26),
    #     dio1=Pin(35),
    #     reset=Pin(14),
    #     lora_cfg=lora_cfg,
    # )
    raise NotImplementedError("Replace this function with one that returns a lora modem instance")


def main():
    print("Initializing...")
    modem = get_modem()

    print("Main loop started")
    receiver = Receiver(modem)

    while True:
        # With wait=True, this function blocks until something is received and always
        # returns non-None
        sender_id, data = receiver.recv(wait=True)

        # Do something with the data!
        print(f"Received {data} from {sender_id:#x}")


class Receiver:
    def __init__(self, modem):
        self.modem = modem
        self.last_counters = {}  # Track the last counter value we got from each sender ID
        self.rx_packet = None  # Reuse RxPacket object when possible, save allocation
        self.ack_buffer = bytearray(ACK_LENGTH)  # reuse the same buffer for ACK packets
        self.skipped_packets = 0  # Counter of skipped packets

        modem.calibrate()

        # Start receiving immediately. We expect the modem to receive continuously
        self.will_irq = modem.start_recv(continuous=True)
        print("Modem initialized and started receive...")

    def recv(self, wait=True):
        # Receive a packet from the sender, including sending an ACK.
        #
        # Returns a tuple of the 16-bit sender id and the sensor data payload.
        #
        # This function should be called very frequently from the main loop (at
        # least every ACK_DELAY_MS milliseconds), to avoid not sending ACKs in time.
        #
        # If 'wait' argument is True (default), the function blocks indefinitely
        # until a packet is received. If False then it will return None
        # if no packet is available.
        #
        # Note that because we called start_recv(continuous=True), the modem
        # will keep receiving on its own - even if when we call send() to
        # send an ACK.
        while True:
            rx = self.modem.poll_recv(rx_packet=self.rx_packet)

            if isinstance(rx, RxPacket):  # value will be True or an RxPacket instance
                decoded = self._handle_rx(rx)
                if decoded:
                    return decoded  # valid LoRa packet and valid for this application

            if not wait:
                return None

            # Otherwise, wait for an IRQ (or have a short sleep) and then poll recv again
            # (receiver is not a low power node, so don't bother with sleep modes.)
            if self.will_irq:
                while not self.modem.irq_triggered():
                    machine.idle()
            else:
                time.sleep_ms(1)

    def _handle_rx(self, rx):
        # Internal function to handle a received packet and either send an ACK
        # and return the sender and the payload, or return None if packet
        # payload is invalid or a duplicate.

        if len(rx) < 5:  # 4 byte header plus 1 byte checksum
            print("Invalid packet length")
            return None

        sender_id, counter, data_len = struct.unpack("<HBB", rx)
        csum = rx[-1]

        if len(rx) != data_len + 5:
            print("Invalid length in payload header")
            return None

        calc_csum = sum(b for b in rx[:-1]) & 0xFF
        if csum != calc_csum:
            print(f"Invalid checksum. calc={calc_csum:#x} received={csum:#x}")
            return None

        # Packet is valid!

        if _DEBUG:
            print(f"RX {data_len} byte message RSSI {rx.rssi} at timestamp {rx.ticks_ms}")

        # Send the ACK
        struct.pack_into(
            "<HHBBb", self.ack_buffer, 0, RECEIVER_ID, sender_id, counter, csum, rx.rssi
        )

        # Time send to start as close to ACK_DELAY_MS after message was received as possible
        tx_at_ms = time.ticks_add(rx.ticks_ms, ACK_DELAY_MS)
        tx_done = self.modem.send(self.ack_buffer, tx_at_ms=tx_at_ms)

        if _DEBUG:
            tx_time = time.ticks_diff(tx_done, tx_at_ms)
            expected = self.modem.get_time_on_air_us(ACK_LENGTH) / 1000
            print(f"ACK TX {tx_at_ms}ms -> {tx_done}ms took {tx_time}ms expected {expected}")

        # Check if the data we received is fresh or stale
        if sender_id not in self.last_counters:
            print(f"New device id {sender_id:#x}")
        elif self.last_counters[sender_id] == counter:
            print(f"Duplicate packet received from {sender_id:#x}")
            return None
        elif counter != 1:
            # If the counter from this sender has gone up by more than 1 since
            # last time we got a packet, we know there is some packet loss.
            #
            # (ignore the case where the new counter is 1, as this probably
            # means a reset.)
            delta = (counter - 1 - self.last_counters[sender_id]) & 0xFF
            if delta:
                print(f"Skipped/lost {delta} packets from {sender_id:#x}")
                self.skipped_packets += delta

        self.last_counters[sender_id] = counter
        return sender_id, rx[4:-1]


if __name__ == "__main__":
    main()



# PEP-0498 specifies that '\\' and '#' must be disallowed explicitly, whereas
# MicroPython relies on the syntax error as a result of the substitution.

print(f"\\")
print(f"#")
try:
    eval("f'{\}'")
except SyntaxError:
    print("SyntaxError")
try:
    eval("f'{#}'")
except SyntaxError:
    print("SyntaxError")



# Test a deep set of "yield from" statements.


def recursive_yield_from(depth, iter_):
    if depth <= 0:
        for i in iter_:
            yield i
    else:
        yield from recursive_yield_from(depth - 1, iter_)


def test(n):
    global result
    result = 0
    for i in recursive_yield_from(10, range(n)):
        result += i


###########################################################################
# Benchmark interface

bm_params = {
    (100, 10): (2000,),
    (1000, 10): (20000,),
    (5000, 10): (100000,),
}


def bm_setup(params):
    (nloop,) = params
    return lambda: test(nloop), lambda: (nloop // 100, result)



# Test of a ESPnow echo server and client transferring encrypted data.
# This test works with ESP32 or ESP8266 as server or client.

# First instance (echo server):
#   Set the shared PMK
#   Set the PEERS global to our mac addresses
#   Run the echo server
#     First exchange an unencrypted message from the client (so we
#     can get its MAC address) and echo the message back (unenecrypted).
#     Then set the peer LMK so all further communications are encrypted.

# Second instance (echo client):
#   Set the shared PMK
#   Send an unencrypted message to the server and wait for echo response.
#   Set the LMK for the peer communications so all further comms are encrypted.
#   Send random messages and compare with response from server.

try:
    import network
    import random
    import time
    import espnow
except ImportError:
    print("SKIP")
    raise SystemExit


# Set read timeout to 5 seconds
timeout_ms = 5000
default_pmk = b"MicroPyth0nRules"
default_lmk = b"0123456789abcdef"
sync = True


def echo_server(e):
    peers = []
    while True:
        # Wait for messages from the client
        peer, msg = e.recv(timeout_ms)
        if peer is None:
            return
        if peer not in peers:
            # If this is first message, add the peer unencrypted
            e.add_peer(peer)

        #  Echo the message back to the sender
        if not e.send(peer, msg, sync):
            print("ERROR: send() failed to", peer)
            return

        if peer not in peers:
            # If this is first message, add the peer encrypted
            peers.append(peer)
            e.del_peer(peer)
            e.add_peer(peer, default_lmk)

        if msg == b"!done":
            return


# Send a message from the client and compare with response from server.
def echo_test(e, peer, msg, sync):
    print("TEST: send/recv(msglen=", len(msg), ",sync=", sync, "): ", end="", sep="")
    try:
        if not e.send(peer, msg, sync):
            print("ERROR: Send failed.")
            return
    except OSError as exc:
        # Don't print exc as it is differs for esp32 and esp8266
        print("ERROR: OSError:")
        return

    p2, msg2 = e.recv(timeout_ms)
    if p2 is None:
        print("ERROR: No response from server.")
        raise SystemExit

    print("OK" if msg2 == msg else "ERROR: Received != Sent")


# Send some random messages to server and check the responses
def echo_client(e, peer, msglens):
    for sync in [True, False]:
        for msglen in msglens:
            msg = bytearray(msglen)
            if msglen > 0:
                msg[0] = b"_"[0]  # Random message must not start with '!'
            for i in range(1, msglen):
                msg[i] = random.getrandbits(8)
            echo_test(e, peer, msg, sync)


# Initialise the wifi and espnow hardware and software
def init(sta_active=True, ap_active=False):
    wlans = [network.WLAN(i) for i in [network.STA_IF, network.AP_IF]]
    e = espnow.ESPNow()
    e.active(True)
    e.set_pmk(default_pmk)
    wlans[0].active(sta_active)
    wlans[1].active(ap_active)
    wlans[0].disconnect()  # Force esp8266 STA interface to disconnect from AP
    return e


# Server
def instance0():
    e = init(True, False)
    macs = [network.WLAN(i).config("mac") for i in (0, 1)]
    print("Server Start")
    multitest.globals(PEERS=macs)
    multitest.next()
    echo_server(e)
    print("Server Done")
    e.active(False)


# Client
def instance1():
    e = init(True, False)
    multitest.next()
    peer = PEERS[0]
    e.add_peer(peer)
    echo_test(e, peer, b"start", True)
    # Wait long enough for the server to set the lmk
    time.sleep(0.1)
    e.del_peer(peer)
    e.add_peer(peer, default_lmk)
    echo_client(e, peer, [250])
    echo_test(e, peer, b"!done", True)
    e.active(False)



# test "soft" machine.Timer (no hardware ID)
import sys


if sys.platform in ("esp32", "esp8266"):
    print("SKIP")  # TODO: Implement soft timers for esp32/esp8266 ports
    raise SystemExit


try:
    import time, machine as machine

    machine.Timer
except:
    print("SKIP")
    raise SystemExit

# create and deinit
t = machine.Timer(freq=1)
t.deinit()

# deinit again
t.deinit()

# create 2 and deinit
t = machine.Timer(freq=1)
t2 = machine.Timer(freq=1)
t.deinit()
t2.deinit()

# create 2 and deinit in different order
t = machine.Timer(freq=1)
t2 = machine.Timer(freq=1)
t2.deinit()
t.deinit()

# create one-shot timer with callback and wait for it to print (should be just once)
t = machine.Timer(period=1, mode=machine.Timer.ONE_SHOT, callback=lambda t: print("one-shot"))
time.sleep_ms(5)
t.deinit()

# create periodic timer with callback and wait for it to print
t = machine.Timer(period=4, mode=machine.Timer.PERIODIC, callback=lambda t: print("periodic"))
time.sleep_ms(14)
t.deinit()



"""
machine test for the CC3200 based boards.
"""

import machine
import os
from network import WLAN

mch = os.uname().machine
if not "LaunchPad" in mch and not "WiPy" in mch:
    raise Exception("Board not supported!")

wifi = WLAN()

print(machine)
machine.idle()
print(machine.freq() == (80000000,))
print(machine.unique_id() == wifi.mac())

machine.main("main.py")

rand_nums = []
for i in range(0, 100):
    rand = machine.rng()
    if rand not in rand_nums:
        rand_nums.append(rand)
    else:
        print("RNG number repeated")
        break

for i in range(0, 10):
    machine.idle()

print("Active")

print(machine.reset_cause() >= 0)
print(machine.wake_reason() >= 0)

try:
    machine.main(123456)
except:
    print("Exception")



# test match.groups()

try:
    import re
except ImportError:
    print("SKIP")
    raise SystemExit

try:
    m = re.match(".", "a")
    m.groups
except AttributeError:
    print("SKIP")
    raise SystemExit


m = re.match(r"(([0-9]*)([a-z]*)[0-9]*)", "1234hello567")
print(m.groups())

m = re.match(r"([0-9]*)(([a-z]*)([0-9]*))", "1234hello567")
print(m.groups())

# optional group that matches
print(re.match(r"(a)?b(c)", "abc").groups())

# optional group that doesn't match
print(re.match(r"(a)?b(c)", "bc").groups())

# only a single match
print(re.match(r"abc", "abc").groups())



import umqtt.robust
import time

# Instantiate an MQTTClient with a keepalive time of 5 seconds (to help us test
# what happens to check_msg() with a broken connection)
m = umqtt.robust.MQTTClient(host="localhost", debug=True, keepalive=5)

m.connect()

# Wait for the broker to consider us dead
time.sleep(6)

# This should initiate a reconnect() and return immediately
m.check_msg()



"""
categories: Modules,os
description: ``getenv`` returns actual value instead of cached value
cause: The ``environ`` attribute is not implemented
workaround: Unknown
"""
import os

print(os.getenv("NEW_VARIABLE"))
os.putenv("NEW_VARIABLE", "VALUE")
print(os.getenv("NEW_VARIABLE"))



# test unary operators


@micropython.viper
def pos(x: int) -> int:
    return +x


print(pos(0))
print(pos(1))
print(pos(-2))


@micropython.viper
def neg(x: int) -> int:
    return -x


print(neg(0))
print(neg(1))
print(neg(-2))


@micropython.viper
def inv(x: int) -> int:
    return ~x


print(inv(0))
print(inv(1))
print(inv(-2))



try:
    from io import StringIO
    import json
except ImportError:
    print("SKIP")
    raise SystemExit

print(json.load(StringIO("null")))
print(json.load(StringIO('"abc\\u0064e"')))
print(json.load(StringIO("[false, true, 1, -2]")))
print(json.load(StringIO('{"a":true}')))



# Checks that an instance type inheriting from a native base that uses
# MP_TYPE_FLAG_ITER_IS_STREAM will still have a getiter.

import io

a = io.StringIO()
a.write("hello\nworld\nmicro\npython\n")
a.seek(0)

for line in a:
    print(line)

class X(io.StringIO):
    pass

b = X()
b.write("hello\nworld\nmicro\npython\n")
b.seek(0)

for line in b:
    print(line)



"""
Define names for built-in types that aren't directly accessible as a builtin.
"""
import sys

# Iterators in Python aren't a matter of type but of protocol.  A large
# and changing number of builtin types implement *some* flavor of
# iterator.  Don't check the type!  Use hasattr to check for both
# "__iter__" and "__next__" attributes instead.


def _f():
    pass


FunctionType = type(_f)
LambdaType = type(lambda: None)  # Same as FunctionType
CodeType = None  # TODO: Add better sentinel which can't match anything
MappingProxyType = None  # TODO: Add better sentinel which can't match anything
SimpleNamespace = None  # TODO: Add better sentinel which can't match anything


def _g():
    yield 1


GeneratorType = type(_g())


class _C:
    def _m(self):
        pass


MethodType = type(_C()._m)

BuiltinFunctionType = type(len)
BuiltinMethodType = type([].append)  # Same as BuiltinFunctionType

ModuleType = type(sys)

try:
    raise TypeError
except TypeError:
    #    tb = sys.exc_info()[2]
    TracebackType = None  # TODO: Add better sentinel which can't match anything
    FrameType = None  # TODO: Add better sentinel which can't match anything
    tb = None
    del tb

# For Jython, the following two types are identical
GetSetDescriptorType = None  # TODO: Add better sentinel which can't match anything
MemberDescriptorType = None  # TODO: Add better sentinel which can't match anything

del (
    sys,
    _f,
    _g,
    _C,
)  # Not for export


# Provide a PEP 3115 compliant mechanism for class creation
def new_class(name, bases=(), kwds=None, exec_body=None):
    """Create a class object dynamically using the appropriate metaclass."""
    meta, ns, kwds = prepare_class(name, bases, kwds)
    if exec_body is not None:
        exec_body(ns)
    return meta(name, bases, ns, **kwds)


def prepare_class(name, bases=(), kwds=None):
    """Call the __prepare__ method of the appropriate metaclass.

    Returns (metaclass, namespace, kwds) as a 3-tuple

    *metaclass* is the appropriate metaclass
    *namespace* is the prepared class namespace
    *kwds* is an updated copy of the passed in kwds argument with any
    'metaclass' entry removed. If no kwds argument is passed in, this will
    be an empty dict.
    """
    if kwds is None:
        kwds = {}
    else:
        kwds = dict(kwds)  # Don't alter the provided mapping
    if "metaclass" in kwds:
        meta = kwds.pop("metaclass")
    else:
        if bases:
            meta = type(bases[0])
        else:
            meta = type
    if isinstance(meta, type):
        # when meta is a type, we first determine the most-derived metaclass
        # instead of invoking the initial candidate directly
        meta = _calculate_meta(meta, bases)
    if hasattr(meta, "__prepare__"):
        ns = meta.__prepare__(name, bases, **kwds)
    else:
        ns = {}
    return meta, ns, kwds


def _calculate_meta(meta, bases):
    """Calculate the most derived metaclass."""
    winner = meta
    for base in bases:
        base_meta = type(base)
        if issubclass(winner, base_meta):
            continue
        if issubclass(base_meta, winner):
            winner = base_meta
            continue
        # else:
        raise TypeError(
            "metaclass conflict: "
            "the metaclass of a derived class "
            "must be a (non-strict) subclass "
            "of the metaclasses of all its bases"
        )
    return winner



try:
    import json
except ImportError:
    print("SKIP")
    raise SystemExit

print(json.dumps(False))
print(json.dumps(True))
print(json.dumps(None))
print(json.dumps(1))
print(json.dumps("abc"))
print(json.dumps("\x00\x01\x7e"))
print(json.dumps([]))
print(json.dumps([1]))
print(json.dumps([1, 2]))
print(json.dumps([1, True]))
print(json.dumps(()))
print(json.dumps((1,)))
print(json.dumps((1, 2)))
print(json.dumps((1, (2, 3))))
print(json.dumps({}))
print(json.dumps({"a": 1}))
print(json.dumps({"a": (2, [3, None])}))
print(json.dumps('"quoted"'))
print(json.dumps("space\n\r\tspace"))
print(json.dumps({None: -1}))
print(json.dumps({False: 0}))
print(json.dumps({True: 1}))
print(json.dumps({1: 2}))



# Testcase for break in a for [within bunch of other code]
# https://github.com/micropython/micropython/issues/635

def foo():
    seq = [1, 2, 3]
    v = 100
    i = 5
    while i > 0:
        print(i)
        for a in seq:
            if a == 2:
                break
        i -= 1

foo()

# break from within nested for loop
def bar():
    l = [1, 2, 3]
    for e1 in l:
        print(e1)
        for e2 in l:
            print(e1, e2)
            if e2 == 2:
                break

bar()



# test errors from bad operations (unary, binary, etc)
try:
    memoryview
except:
    print("SKIP")
    raise SystemExit

# unsupported binary operators
try:
    memoryview(b"") + b""
except TypeError:
    print("TypeError")

try:
    memoryview(b"") + memoryview(b"")
except TypeError:
    print("TypeError")

try:
    m = memoryview(bytearray())
    m += bytearray()
except TypeError:
    print("TypeError")



# test construction of bytes from different objects
try:
    from array import array
except ImportError:
    print("SKIP")
    raise SystemExit

# arrays
print(bytes(array('b', [1, 2])))
print(bytes(array('h', [0x101, 0x202])))



HIGHEST_PROTOCOL = 0


def dump(obj, f, proto=0):
    f.write(repr(obj))


def dumps(obj, proto=0):
    return repr(obj).encode()


def load(f):
    s = f.read()
    return loads(s)


def loads(s):
    d = {}
    s = s.decode()
    if "(" in s:
        qualname = s.split("(", 1)[0]
        if "." in qualname:
            pkg = qualname.rsplit(".", 1)[0]
            mod = __import__(pkg)
            d[pkg] = mod
    return eval(s, d)



def count(start=0, step=1):
    while True:
        yield start
        start += step


def cycle(p):
    try:
        len(p)
    except TypeError:
        # len() is not defined for this type. Assume it is
        # a finite iterable so we must cache the elements.
        cache = []
        for i in p:
            yield i
            cache.append(i)
        p = cache
    while p:
        yield from p


def repeat(el, n=None):
    if n is None:
        while True:
            yield el
    else:
        for i in range(n):
            yield el


def chain(*p):
    for i in p:
        yield from i


def islice(p, start, stop=(), step=1):
    if stop == ():
        stop = start
        start = 0
    # TODO: optimizing or breaking semantics?
    if start >= stop:
        return
    it = iter(p)
    for i in range(start):
        next(it)

    while True:
        yield next(it)
        for i in range(step - 1):
            next(it)
        start += step
        if start >= stop:
            return


def tee(iterable, n=2):
    return [iter(iterable)] * n


def starmap(function, iterable):
    for args in iterable:
        yield function(*args)


def accumulate(iterable, func=lambda x, y: x + y):
    it = iter(iterable)
    try:
        acc = next(it)
    except StopIteration:
        return
    yield acc
    for element in it:
        acc = func(acc, element)
        yield acc



from machine import Pin, Signal

# 96Boards Carbon board
# USR1 - User controlled led, connected to PD2
# USR2 - User controlled led, connected to PA15
# BT - Bluetooth indicator, connected to PB5.
# Note - 96b_carbon uses (at the time of writing) non-standard
# for Zephyr port device naming convention.
LED = Signal(("GPIOA", 15), Pin.OUT)



# This module provides a backwards-compatble import for `urequests`.
# It lazy-loads from `requests` without duplicating its globals dict.


def __getattr__(attr):
    import requests

    return getattr(requests, attr)



try:
    import binascii
except ImportError:
    print("SKIP")
    raise SystemExit

print(binascii.b2a_base64(b""))
print(binascii.b2a_base64(b"f"))
print(binascii.b2a_base64(b"fo"))
print(binascii.b2a_base64(b"foo"))
print(binascii.b2a_base64(b"foob"))
print(binascii.b2a_base64(b"fooba"))
print(binascii.b2a_base64(b"foobar"))

print(binascii.b2a_base64(b"\x00\x01\x02\x03\x04\x05\x06\x07"))
print(binascii.b2a_base64(b"\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"))
print(binascii.b2a_base64(b"\x7f\x80\xff"))
print(binascii.b2a_base64(b"1234ABCDabcd"))
print(binascii.b2a_base64(b"\x00\x00>"))  # convert into '+'

print(binascii.b2a_base64(b"foobar", newline=True))
print(binascii.b2a_base64(b"foobar", newline=False))



try:
    import os, vfs

    vfs.VfsFat
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class RAMFS:
    SEC_SIZE = 512

    def __init__(self, blocks):
        self.data = bytearray(blocks * self.SEC_SIZE)

    def readblocks(self, n, buf):
        # print("readblocks(%s, %x(%d))" % (n, id(buf), len(buf)))
        for i in range(len(buf)):
            buf[i] = self.data[n * self.SEC_SIZE + i]

    def writeblocks(self, n, buf):
        # print("writeblocks(%s, %x)" % (n, id(buf)))
        for i in range(len(buf)):
            self.data[n * self.SEC_SIZE + i] = buf[i]

    def ioctl(self, op, arg):
        # print("ioctl(%d, %r)" % (op, arg))
        if op == 4:  # MP_BLOCKDEV_IOCTL_BLOCK_COUNT
            return len(self.data) // self.SEC_SIZE
        if op == 5:  # MP_BLOCKDEV_IOCTL_BLOCK_SIZE
            return self.SEC_SIZE


try:
    bdev = RAMFS(50)
    bdev2 = RAMFS(50)
except MemoryError:
    print("SKIP")
    raise SystemExit

# first we umount any existing mount points the target may have
try:
    vfs.umount("/")
except OSError:
    pass
for path in os.listdir("/"):
    vfs.umount("/" + path)

vfs.VfsFat.mkfs(bdev)
vfs.mount(bdev, "/")

print(os.getcwd())

f = open("test.txt", "w")
f.write("hello")
f.close()

print(os.listdir())
print(os.listdir("/"))
print(os.stat("")[:-3])
print(os.stat("/")[:-3])
print(os.stat("test.txt")[:-3])
print(os.stat("/test.txt")[:-3])

f = open("/test.txt")
print(f.read())
f.close()

os.rename("test.txt", "test2.txt")
print(os.listdir())
os.rename("test2.txt", "/test3.txt")
print(os.listdir())
os.rename("/test3.txt", "test4.txt")
print(os.listdir())
os.rename("/test4.txt", "/test5.txt")
print(os.listdir())

os.mkdir("dir")
print(os.listdir())
os.mkdir("/dir2")
print(os.listdir())
os.mkdir("dir/subdir")
print(os.listdir("dir"))
for exist in ("", "/", "dir", "/dir", "dir/subdir"):
    try:
        os.mkdir(exist)
    except OSError as er:
        print("mkdir OSError", er.errno == 17)  # EEXIST

os.chdir("/")
print(os.stat("test5.txt")[:-3])

vfs.VfsFat.mkfs(bdev2)
vfs.mount(bdev2, "/sys")
print(os.listdir())
print(os.listdir("sys"))
print(os.listdir("/sys"))

os.rmdir("dir2")
os.remove("test5.txt")
print(os.listdir())

vfs.umount("/")
print(os.getcwd())
print(os.listdir())
print(os.listdir("sys"))

# test importing a file from a mounted FS
import sys

sys.path.clear()
sys.path.append("/sys")
with open("sys/test_module.py", "w") as f:
    f.write('print("test_module!")')
import test_module



# Basic MAX17048 library for OMGS3 and other Unexpected Maker products
# MIT license; Copyright (c) 2024 Seon Rozenblum - Unexpected Maker
#
# Project home:
#   https://unexpectedmaker.com

from machine import I2C


class MAX17048:
    _MAX17048_ADDRESS = 0x36

    _VCELL_REGISTER = 0x02
    _SOC_REGISTER = 0x04
    _MODE_REGISTER = 0x06
    _VERSION_REGISTER = 0x08
    _HIBRT_REGISTER = 0x0A
    _CONFIG_REGISTER = 0x0C
    _COMMAND_REGISTER = 0xFE

    def __init__(self, i2c, address=_MAX17048_ADDRESS):
        self.i2c = i2c
        self.address = address

    def _read_register(self, register, num_bytes):
        result = self.i2c.readfrom_mem(self.address, register, num_bytes)
        return int.from_bytes(result, "big")

    def _write_register(self, register, value, num_bytes):
        data = value.to_bytes(num_bytes, "big")
        self.i2c.writeto_mem(self.address, register, data)

    @property
    def cell_voltage(self):
        """The voltage of the connected cell in Volts."""
        raw_voltage = self._read_register(self._VCELL_REGISTER, 2)
        voltage = (raw_voltage >> 4) * 0.00125
        return voltage

    @property
    def state_of_charge(self):
        """The state of charge of the battery in percentage."""
        raw_soc = self._read_register(self._SOC_REGISTER, 2)
        return raw_soc / 256

    @property
    def version(self):
        """The chip version."""
        return self._read_register(self._VERSION_REGISTER, 2)

    @property
    def hibernate(self):
        """True if the chip is in hibernate mode, False otherwise."""
        hib = self._read_register(self._HIBRT_REGISTER, 2)
        return (hib & 0x4000) != 0

    @hibernate.setter
    def hibernate(self, value):
        config = self._read_register(self._CONFIG_REGISTER, 2)
        if value:
            config |= 0x8000  # Set the sleep bit
        else:
            config &= ~0x8000  # Clear the sleep bit
        self._write_register(self._CONFIG_REGISTER, config, 2)

    def quick_start(self):
        """Perform a quick start to reset the SOC calculation in the chip."""
        self._write_register(self._MODE_REGISTER, 0x4000, 2)

    def reset(self):
        """Reset the chip."""
        self._write_register(self._COMMAND_REGISTER, 0x5400, 2)



"""
Driver for accelerometer on STM32F4 Discover board.

Sets accelerometer range at +-2g.
Returns list containing X,Y,Z axis acceleration values in 'g' units (9.8m/s^2).

See:
    STM32Cube_FW_F4_V1.1.0/Drivers/BSP/Components/lis302dl/lis302dl.h
    STM32Cube_FW_F4_V1.1.0/Drivers/BSP/Components/lis302dl/lis302dl.c
    STM32Cube_FW_F4_V1.1.0/Drivers/BSP/STM32F4-Discovery/stm32f4_discovery.c
    STM32Cube_FW_F4_V1.1.0/Drivers/BSP/STM32F4-Discovery/stm32f4_discovery.h
    STM32Cube_FW_F4_V1.1.0/Drivers/BSP/STM32F4-Discovery/stm32f4_discovery_accelerometer.c
    STM32Cube_FW_F4_V1.1.0/Drivers/BSP/STM32F4-Discovery/stm32f4_discovery_accelerometer.h
    STM32Cube_FW_F4_V1.1.0/Projects/STM32F4-Discovery/Demonstrations/Src/main.c
"""

from micropython import const
from pyb import Pin
from pyb import SPI

READWRITE_CMD = const(0x80)
MULTIPLEBYTE_CMD = const(0x40)
WHO_AM_I_ADDR = const(0x0F)
OUT_X_ADDR = const(0x29)
OUT_Y_ADDR = const(0x2B)
OUT_Z_ADDR = const(0x2D)
OUT_T_ADDR = const(0x0C)

LIS302DL_WHO_AM_I_VAL = const(0x3B)
LIS302DL_CTRL_REG1_ADDR = const(0x20)
# Configuration for 100Hz sampling rate, +-2g range
LIS302DL_CONF = const(0b01000111)

LIS3DSH_WHO_AM_I_VAL = const(0x3F)
LIS3DSH_CTRL_REG4_ADDR = const(0x20)
LIS3DSH_CTRL_REG5_ADDR = const(0x24)
# Configuration for 100Hz sampling rate, +-2g range
LIS3DSH_CTRL_REG4_CONF = const(0b01100111)
LIS3DSH_CTRL_REG5_CONF = const(0b00000000)


class STAccel:
    def __init__(self):
        self.cs_pin = Pin("PE3", Pin.OUT_PP, Pin.PULL_NONE)
        self.cs_pin.high()
        self.spi = SPI(1, SPI.MASTER, baudrate=328125, polarity=0, phase=1, bits=8)

        self.who_am_i = self.read_id()

        if self.who_am_i == LIS302DL_WHO_AM_I_VAL:
            self.write_bytes(LIS302DL_CTRL_REG1_ADDR, bytearray([LIS302DL_CONF]))
            self.sensitivity = 18
        elif self.who_am_i == LIS3DSH_WHO_AM_I_VAL:
            self.write_bytes(LIS3DSH_CTRL_REG4_ADDR, bytearray([LIS3DSH_CTRL_REG4_CONF]))
            self.write_bytes(LIS3DSH_CTRL_REG5_ADDR, bytearray([LIS3DSH_CTRL_REG5_CONF]))
            self.sensitivity = 0.06 * 256
        else:
            raise Exception("LIS302DL or LIS3DSH accelerometer not present")

    def convert_raw_to_g(self, x):
        if x & 0x80:
            x = x - 256
        return x * self.sensitivity / 1000

    def read_bytes(self, addr, nbytes):
        if nbytes > 1:
            addr |= READWRITE_CMD | MULTIPLEBYTE_CMD
        else:
            addr |= READWRITE_CMD
        self.cs_pin.low()
        self.spi.send(addr)
        # buf = self.spi.send_recv(bytearray(nbytes * [0])) # read data, MSB first
        buf = self.spi.recv(nbytes)
        self.cs_pin.high()
        return buf

    def write_bytes(self, addr, buf):
        if len(buf) > 1:
            addr |= MULTIPLEBYTE_CMD
        self.cs_pin.low()
        self.spi.send(addr)
        for b in buf:
            self.spi.send(b)
        self.cs_pin.high()

    def read_id(self):
        return self.read_bytes(WHO_AM_I_ADDR, 1)[0]

    def x(self):
        return self.convert_raw_to_g(self.read_bytes(OUT_X_ADDR, 1)[0])

    def y(self):
        return self.convert_raw_to_g(self.read_bytes(OUT_Y_ADDR, 1)[0])

    def z(self):
        return self.convert_raw_to_g(self.read_bytes(OUT_Z_ADDR, 1)[0])

    def xyz(self):
        return (self.x(), self.y(), self.z())



"""Test for nrf24l01 module.  Portable between MicroPython targets."""

import sys
import struct
import utime
from machine import Pin, SPI, SoftSPI
from nrf24l01 import NRF24L01
from micropython import const

# Responder pause between receiving data and checking for further packets.
_RX_POLL_DELAY = const(15)
# Responder pauses an additional _RESPONER_SEND_DELAY ms after receiving data and before
# transmitting to allow the (remote) initiator time to get into receive mode. The
# initiator may be a slow device. Value tested with Pyboard, ESP32 and ESP8266.
_RESPONDER_SEND_DELAY = const(10)

if sys.platform == "pyboard":
    spi = SPI(2)  # miso : Y7, mosi : Y8, sck : Y6
    cfg = {"spi": spi, "csn": "Y5", "ce": "Y4"}
elif sys.platform == "esp8266":  # Hardware SPI
    spi = SPI(1)  # miso : 12, mosi : 13, sck : 14
    cfg = {"spi": spi, "csn": 4, "ce": 5}
elif sys.platform == "esp32":  # Software SPI
    spi = SoftSPI(sck=Pin(25), mosi=Pin(33), miso=Pin(32))
    cfg = {"spi": spi, "csn": 26, "ce": 27}
elif sys.platform == "rp2":  # Hardware SPI with explicit pin definitions
    spi = SPI(0, sck=Pin(2), mosi=Pin(3), miso=Pin(4))
    cfg = {"spi": spi, "csn": 5, "ce": 6}
else:
    raise ValueError("Unsupported platform {}".format(sys.platform))

# Addresses are in little-endian format. They correspond to big-endian
# 0xf0f0f0f0e1, 0xf0f0f0f0d2
pipes = (b"\xe1\xf0\xf0\xf0\xf0", b"\xd2\xf0\xf0\xf0\xf0")


def initiator():
    csn = Pin(cfg["csn"], mode=Pin.OUT, value=1)
    ce = Pin(cfg["ce"], mode=Pin.OUT, value=0)
    spi = cfg["spi"]
    nrf = NRF24L01(spi, csn, ce, payload_size=8)

    nrf.open_tx_pipe(pipes[0])
    nrf.open_rx_pipe(1, pipes[1])
    nrf.start_listening()

    num_needed = 16
    num_successes = 0
    num_failures = 0
    led_state = 0

    print("NRF24L01 initiator mode, sending %d packets..." % num_needed)

    while num_successes < num_needed and num_failures < num_needed:
        # stop listening and send packet
        nrf.stop_listening()
        millis = utime.ticks_ms()
        led_state = max(1, (led_state << 1) & 0x0F)
        print("sending:", millis, led_state)
        try:
            nrf.send(struct.pack("ii", millis, led_state))
        except OSError:
            pass

        # start listening again
        nrf.start_listening()

        # wait for response, with 250ms timeout
        start_time = utime.ticks_ms()
        timeout = False
        while not nrf.any() and not timeout:
            if utime.ticks_diff(utime.ticks_ms(), start_time) > 250:
                timeout = True

        if timeout:
            print("failed, response timed out")
            num_failures += 1

        else:
            # recv packet
            (got_millis,) = struct.unpack("i", nrf.recv())

            # print response and round-trip delay
            print(
                "got response:",
                got_millis,
                "(delay",
                utime.ticks_diff(utime.ticks_ms(), got_millis),
                "ms)",
            )
            num_successes += 1

        # delay then loop
        utime.sleep_ms(250)

    print("initiator finished sending; successes=%d, failures=%d" % (num_successes, num_failures))


def responder():
    csn = Pin(cfg["csn"], mode=Pin.OUT, value=1)
    ce = Pin(cfg["ce"], mode=Pin.OUT, value=0)
    spi = cfg["spi"]
    nrf = NRF24L01(spi, csn, ce, payload_size=8)

    nrf.open_tx_pipe(pipes[1])
    nrf.open_rx_pipe(1, pipes[0])
    nrf.start_listening()

    print("NRF24L01 responder mode, waiting for packets... (ctrl-C to stop)")

    while True:
        if nrf.any():
            while nrf.any():
                buf = nrf.recv()
                millis, led_state = struct.unpack("ii", buf)
                print("received:", millis, led_state)
                for led in leds:
                    if led_state & 1:
                        led.on()
                    else:
                        led.off()
                    led_state >>= 1
                utime.sleep_ms(_RX_POLL_DELAY)

            # Give initiator time to get into receive mode.
            utime.sleep_ms(_RESPONDER_SEND_DELAY)
            nrf.stop_listening()
            try:
                nrf.send(struct.pack("i", millis))
            except OSError:
                pass
            print("sent response")
            nrf.start_listening()


try:
    import pyb

    leds = [pyb.LED(i + 1) for i in range(4)]
except:
    leds = []

print("NRF24L01 test module loaded")
print("NRF24L01 pinout for test:")
print("    CE on", cfg["ce"])
print("    CSN on", cfg["csn"])
print("    SPI on", cfg["spi"])
print("run nrf24l01test.responder() on responder, then nrf24l01test.initiator() on initiator")



import errno
import os

from micropython import const

_SEP = const("/")


def _mode_if_exists(path):
    try:
        return os.stat(path)[0]
    except OSError as e:
        if e.errno == errno.ENOENT:
            return 0
        raise e


def _clean_segment(segment):
    segment = str(segment)
    if not segment:
        return "."
    segment = segment.rstrip(_SEP)
    if not segment:
        return _SEP
    while True:
        no_double = segment.replace(_SEP + _SEP, _SEP)
        if no_double == segment:
            break
        segment = no_double
    return segment


class Path:
    def __init__(self, *segments):
        segments_cleaned = []
        for segment in segments:
            segment = _clean_segment(segment)
            if segment[0] == _SEP:
                segments_cleaned = [segment]
            elif segment == ".":
                continue
            else:
                segments_cleaned.append(segment)

        self._path = _clean_segment(_SEP.join(segments_cleaned))

    def __truediv__(self, other):
        return Path(self._path, str(other))

    def __rtruediv__(self, other):
        return Path(other, self._path)

    def __repr__(self):
        return f'{type(self).__name__}("{self._path}")'

    def __str__(self):
        return self._path

    def __eq__(self, other):
        return self.absolute() == Path(other).absolute()

    def absolute(self):
        path = self._path
        cwd = os.getcwd()
        if not path or path == ".":
            return cwd
        if path[0] == _SEP:
            return path
        return _SEP + path if cwd == _SEP else cwd + _SEP + path

    def resolve(self):
        return self.absolute()

    def open(self, mode="r", encoding=None):
        return open(self._path, mode, encoding=encoding)

    def exists(self):
        return bool(_mode_if_exists(self._path))

    def mkdir(self, parents=False, exist_ok=False):
        try:
            os.mkdir(self._path)
            return
        except OSError as e:
            if e.errno == errno.EEXIST and exist_ok:
                return
            elif e.errno == errno.ENOENT and parents:
                pass  # handled below
            else:
                raise e

        segments = self._path.split(_SEP)
        progressive_path = ""
        if segments[0] == "":
            segments = segments[1:]
            progressive_path = _SEP
        for segment in segments:
            progressive_path += _SEP + segment
            try:
                os.mkdir(progressive_path)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise e

    def is_dir(self):
        return bool(_mode_if_exists(self._path) & 0x4000)

    def is_file(self):
        return bool(_mode_if_exists(self._path) & 0x8000)

    def _glob(self, path, pattern, recursive):
        # Currently only supports a single "*" pattern.
        n_wildcards = pattern.count("*")
        n_single_wildcards = pattern.count("?")

        if n_single_wildcards:
            raise NotImplementedError("? single wildcards not implemented.")

        if n_wildcards == 0:
            raise ValueError
        elif n_wildcards > 1:
            raise NotImplementedError("Multiple * wildcards not implemented.")

        prefix, suffix = pattern.split("*")

        for name, mode, *_ in os.ilistdir(path):
            full_path = path + _SEP + name
            if name.startswith(prefix) and name.endswith(suffix):
                yield full_path
            if recursive and mode & 0x4000:  # is_dir
                yield from self._glob(full_path, pattern, recursive=recursive)

    def glob(self, pattern):
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.

        Currently only supports a single "*" pattern.
        """
        return self._glob(self._path, pattern, recursive=False)

    def rglob(self, pattern):
        return self._glob(self._path, pattern, recursive=True)

    def stat(self):
        return os.stat(self._path)

    def read_bytes(self):
        with open(self._path, "rb") as f:
            return f.read()

    def read_text(self, encoding=None):
        with open(self._path, "r", encoding=encoding) as f:
            return f.read()

    def rename(self, target):
        os.rename(self._path, target)

    def rmdir(self):
        os.rmdir(self._path)

    def touch(self, exist_ok=True):
        if self.exists():
            if exist_ok:
                return  # TODO: should update timestamp
            else:
                # In lieue of FileExistsError
                raise OSError(errno.EEXIST)
        with open(self._path, "w"):
            pass

    def unlink(self, missing_ok=False):
        try:
            os.unlink(self._path)
        except OSError as e:
            if not (missing_ok and e.errno == errno.ENOENT):
                raise e

    def write_bytes(self, data):
        with open(self._path, "wb") as f:
            f.write(data)

    def write_text(self, data, encoding=None):
        with open(self._path, "w", encoding=encoding) as f:
            f.write(data)

    def with_suffix(self, suffix):
        index = -len(self.suffix) or None
        return Path(self._path[:index] + suffix)

    @property
    def stem(self):
        return self.name.rsplit(".", 1)[0]

    @property
    def parent(self):
        tokens = self._path.rsplit(_SEP, 1)
        if len(tokens) == 2:
            if not tokens[0]:
                tokens[0] = _SEP
            return Path(tokens[0])
        return Path(".")

    @property
    def name(self):
        return self._path.rsplit(_SEP, 1)[-1]

    @property
    def suffix(self):
        elems = self._path.rsplit(".", 1)
        return "" if len(elems) == 1 else "." + elems[1]



# test using an OrderedDict as the locals to construct a class

try:
    from collections import OrderedDict
except ImportError:
    print("SKIP")
    raise SystemExit

if not hasattr(int, "__dict__"):
    print("SKIP")
    raise SystemExit


A = type("A", (), OrderedDict(a=1, b=2, c=3, d=4, e=5))
print([k for k in A.__dict__.keys() if not k.startswith("_")])



from pyb import Pin, ADCAll

pins = [Pin.cpu.A0, Pin.cpu.A1, Pin.cpu.A2, Pin.cpu.A3]

# set pins to IN mode, init ADCAll, then check pins are ANALOG
for p in pins:
    p.init(p.IN)
adc = ADCAll(12)
for p in pins:
    print(p)

# set pins to IN mode, init ADCAll with mask, then check some pins are ANALOG
for p in pins:
    p.init(p.IN)
adc = ADCAll(12, 0x70003)
for p in pins:
    print(p)

# init all pins to ANALOG
adc = ADCAll(12)
print(adc)

# read all channels
for c in range(19):
    print(type(adc.read_channel(c)))

# call special reading functions
print(0 < adc.read_core_temp() < 100)
print(0 < adc.read_core_vbat() < 4)
print(0 < adc.read_core_vref() < 2)
print(0 < adc.read_vref() < 4)



# _boot.py
# Try to mount the filesystem, and format the flash if it doesn't exist.
# Note: the flash requires the programming size to be aligned to 256 bytes.

import os
import vfs
import sys
import mimxrt
from machine import Pin

bdev = mimxrt.Flash()
try:
    fs = vfs.VfsLfs2(bdev, progsize=256)
except:
    vfs.VfsLfs2.mkfs(bdev, progsize=256)
    fs = vfs.VfsLfs2(bdev, progsize=256)
vfs.mount(fs, "/flash")
os.chdir("/flash")
sys.path.append("/flash")
sys.path.append("/flash/lib")

# do not mount the SD card if SKIPSD exists.
try:
    os.stat("SKIPSD")
except:
    try:
        from machine import SDCard

        sdcard = SDCard(1)

        fat = vfs.VfsFat(sdcard)
        vfs.mount(fat, "/sdcard")
        os.chdir("/sdcard")
        sys.path.append("/sdcard")
    except:
        pass  # Fail silently



# test waiting within "async for" __anext__ function

import sys
if sys.implementation.name == 'micropython':
    # uPy allows normal generators to be awaitables
    coroutine = lambda f: f
else:
    import types
    coroutine = types.coroutine

@coroutine
def f(x):
    print('f start:', x)
    yield x + 1
    yield x + 2
    return x + 3

class ARange:
    def __init__(self, high):
        print('init')
        self.cur = 0
        self.high = high

    def __aiter__(self):
        print('aiter')
        return self

    async def __anext__(self):
        print('anext')
        print('f returned:', await f(20))
        if self.cur < self.high:
            val = self.cur
            self.cur += 1
            return val
        else:
            raise StopAsyncIteration

async def coro():
    async for x in ARange(4):
        print('x', x)

o = coro()
try:
    while True:
        print('coro yielded:', o.send(None))
except StopIteration:
    print('finished')



# MicroPython asyncio module
# MIT license; Copyright (c) 2019-2020 Damien P. George

from . import core


# Lock class for primitive mutex capability
class Lock:
    def __init__(self):
        # The state can take the following values:
        # - 0: unlocked
        # - 1: locked
        # - <Task>: unlocked but this task has been scheduled to acquire the lock next
        self.state = 0
        # Queue of Tasks waiting to acquire this Lock
        self.waiting = core.TaskQueue()

    def locked(self):
        return self.state == 1

    def release(self):
        if self.state != 1:
            raise RuntimeError("Lock not acquired")
        if self.waiting.peek():
            # Task(s) waiting on lock, schedule next Task
            self.state = self.waiting.pop()
            core._task_queue.push(self.state)
        else:
            # No Task waiting so unlock
            self.state = 0

    # async
    def acquire(self):
        if self.state != 0:
            # Lock unavailable, put the calling Task on the waiting queue
            self.waiting.push(core.cur_task)
            # Set calling task's data to the lock's queue so it can be removed if needed
            core.cur_task.data = self.waiting
            try:
                yield
            except core.CancelledError as er:
                if self.state == core.cur_task:
                    # Cancelled while pending on resume, schedule next waiting Task
                    self.state = 1
                    self.release()
                raise er
        # Lock available, set it as locked
        self.state = 1
        return True

    async def __aenter__(self):
        return await self.acquire()

    async def __aexit__(self, exc_type, exc, tb):
        return self.release()



"""Constants and membership tests for ASCII characters"""

NUL = 0x00  # ^@
SOH = 0x01  # ^A
STX = 0x02  # ^B
ETX = 0x03  # ^C
EOT = 0x04  # ^D
ENQ = 0x05  # ^E
ACK = 0x06  # ^F
BEL = 0x07  # ^G
BS = 0x08  # ^H
TAB = 0x09  # ^I
HT = 0x09  # ^I
LF = 0x0A  # ^J
NL = 0x0A  # ^J
VT = 0x0B  # ^K
FF = 0x0C  # ^L
CR = 0x0D  # ^M
SO = 0x0E  # ^N
SI = 0x0F  # ^O
DLE = 0x10  # ^P
DC1 = 0x11  # ^Q
DC2 = 0x12  # ^R
DC3 = 0x13  # ^S
DC4 = 0x14  # ^T
NAK = 0x15  # ^U
SYN = 0x16  # ^V
ETB = 0x17  # ^W
CAN = 0x18  # ^X
EM = 0x19  # ^Y
SUB = 0x1A  # ^Z
ESC = 0x1B  # ^[
FS = 0x1C  # ^\
GS = 0x1D  # ^]
RS = 0x1E  # ^^
US = 0x1F  # ^_
SP = 0x20  # space
DEL = 0x7F  # delete

controlnames = [
    "NUL",
    "SOH",
    "STX",
    "ETX",
    "EOT",
    "ENQ",
    "ACK",
    "BEL",
    "BS",
    "HT",
    "LF",
    "VT",
    "FF",
    "CR",
    "SO",
    "SI",
    "DLE",
    "DC1",
    "DC2",
    "DC3",
    "DC4",
    "NAK",
    "SYN",
    "ETB",
    "CAN",
    "EM",
    "SUB",
    "ESC",
    "FS",
    "GS",
    "RS",
    "US",
    "SP",
]


def _ctoi(c):
    if type(c) == type(""):
        return ord(c)
    else:
        return c


def isalnum(c):
    return isalpha(c) or isdigit(c)


def isalpha(c):
    return isupper(c) or islower(c)


def isascii(c):
    return _ctoi(c) <= 127  # ?


def isblank(c):
    return _ctoi(c) in (8, 32)


def iscntrl(c):
    return _ctoi(c) <= 31


def isdigit(c):
    return _ctoi(c) >= 48 and _ctoi(c) <= 57


def isgraph(c):
    return _ctoi(c) >= 33 and _ctoi(c) <= 126


def islower(c):
    return _ctoi(c) >= 97 and _ctoi(c) <= 122


def isprint(c):
    return _ctoi(c) >= 32 and _ctoi(c) <= 126


def ispunct(c):
    return _ctoi(c) != 32 and not isalnum(c)


def isspace(c):
    return _ctoi(c) in (9, 10, 11, 12, 13, 32)


def isupper(c):
    return _ctoi(c) >= 65 and _ctoi(c) <= 90


def isxdigit(c):
    return (
        isdigit(c) or (_ctoi(c) >= 65 and _ctoi(c) <= 70) or (_ctoi(c) >= 97 and _ctoi(c) <= 102)
    )


def isctrl(c):
    return _ctoi(c) < 32


def ismeta(c):
    return _ctoi(c) > 127


def ascii(c):
    if type(c) == type(""):
        return chr(_ctoi(c) & 0x7F)
    else:
        return _ctoi(c) & 0x7F


def ctrl(c):
    if type(c) == type(""):
        return chr(_ctoi(c) & 0x1F)
    else:
        return _ctoi(c) & 0x1F


def alt(c):
    if type(c) == type(""):
        return chr(_ctoi(c) | 0x80)
    else:
        return _ctoi(c) | 0x80


def unctrl(c):
    bits = _ctoi(c)
    if bits == 0x7F:
        rep = "^?"
    elif isprint(bits & 0x7F):
        rep = chr(bits & 0x7F)
    else:
        rep = "^" + chr(((bits & 0x7F) | 0x20) + 0x20)
    if bits & 0x80:
        return "!" + rep
    return rep



try:
    import uctypes
except ImportError:
    print("SKIP")
    raise SystemExit

desc = {
    "f1": 0 | uctypes.UINT32,
    "f2": 4 | uctypes.UINT8,
}


# uctypes.NATIVE is default
print(uctypes.sizeof(desc) == uctypes.sizeof(desc, uctypes.NATIVE))

# Here we assume that that we run on a platform with conventional ABI
# (which rounds up structure size based on max alignment). For platforms
# where that doesn't hold, this tests should be just disabled in the runner.
print(uctypes.sizeof(desc, uctypes.NATIVE) > uctypes.sizeof(desc, uctypes.LITTLE_ENDIAN))

# When taking sizeof of instantiated structure, layout type param
# is prohibited (because structure already has its layout type).
s = uctypes.struct(0, desc, uctypes.LITTLE_ENDIAN)
try:
    uctypes.sizeof(s, uctypes.LITTLE_ENDIAN)
except TypeError:
    print("TypeError")



# MicroPython zlib module
# MIT license; Copyright (c) 2023 Jim Mussared

import io, deflate

_MAX_WBITS = const(15)


def _decode_wbits(wbits, decompress):
    if -15 <= wbits <= -5:
        return (
            deflate.RAW,
            -wbits,
        )
    elif 5 <= wbits <= 15:
        return (deflate.ZLIB, wbits)
    elif decompress and wbits == 0:
        return (deflate.ZLIB,)
    elif 21 <= wbits <= 31:
        return (deflate.GZIP, wbits - 16)
    elif decompress and 35 <= wbits <= 47:
        return (deflate.AUTO, wbits - 32)
    else:
        raise ValueError("wbits")


if hasattr(deflate.DeflateIO, "write"):

    def compress(data, wbits=_MAX_WBITS):
        f = io.BytesIO()
        with deflate.DeflateIO(f, *_decode_wbits(wbits, False)) as g:
            g.write(data)
        return f.getvalue()


def decompress(data, wbits=_MAX_WBITS):
    f = io.BytesIO(data)
    with deflate.DeflateIO(f, *_decode_wbits(wbits, True)) as g:
        return g.read()



import itertools

assert list(itertools.islice(list(range(10)), 4)) == [0, 1, 2, 3]
assert list(itertools.islice(list(range(10)), 2, 6)) == [2, 3, 4, 5]
assert list(itertools.islice(list(range(10)), 2, 6, 2)) == [2, 4]


def g():
    while True:
        yield 123


assert list(itertools.islice(g(), 5)) == [123, 123, 123, 123, 123]

assert list(itertools.islice(itertools.cycle([1, 2, 3]), 10)) == [1, 2, 3, 1, 2, 3, 1, 2, 3, 1]
assert list(itertools.islice(itertools.cycle(reversed([1, 2, 3])), 7)) == [3, 2, 1, 3, 2, 1, 3]

assert list(itertools.starmap(lambda x, y: x * y, [[1, 2], [2, 3], [3, 4]])) == [2, 6, 12]

assert list(itertools.accumulate([])) == []
assert list(itertools.accumulate([0])) == [0]
assert list(itertools.accumulate([0, 2, 3])) == [0, 2, 5]
assert list(itertools.accumulate(reversed([0, 2, 3]))) == [3, 5, 5]
assert list(itertools.accumulate([1, 2, 3], lambda x, y: x * y)) == [1, 2, 6]



# Application firmware update function for LEGO_HUB_NO6.
# MIT license; Copyright (c) 2022-2024 Damien P. George

from micropython import const
import random, struct, machine, fwupdate, spiflash, pyb

_IOCTL_BLOCK_COUNT = const(4)
_IOCTL_BLOCK_SIZE = const(5)

# Mboot addresses the external SPI flash at this location.
_MBOOT_SPIFLASH_BASE_ADDR = 0x80000000

# The raw filesystem is in the first 1MiB of external SPI flash,
# but skip the first and last flash sectors.
_RAW_FILESYSTEM_ADDR = const(4 * 1024)
_RAW_FILESYSTEM_LEN = const(1016 * 1024)


def _copy_file_to_raw_filesystem(filename, flash, block):
    block_count = flash.ioctl(_IOCTL_BLOCK_COUNT, 0)
    block_size = flash.ioctl(_IOCTL_BLOCK_SIZE, 0)
    buf = bytearray(block_size)
    with open(filename, "rb") as file:
        while True:
            n = file.readinto(buf)
            if not n:
                break
            flash.writeblocks(block, buf)
            block += 1
            if block >= block_count:
                print("|", end="")
                block = 0
            print(".", end="")
    print()


def update_app(filename):
    print(f"Updating application firmware from {filename}")

    # We can't use pyb.Flash() because we need to write to the "reserved" 1M area.
    flash = spiflash.SPIFlash(start=_RAW_FILESYSTEM_ADDR, len=_RAW_FILESYSTEM_LEN)

    # Partition the raw filesystem into two segments for wear levelling.
    block_count = flash.ioctl(_IOCTL_BLOCK_COUNT, 0)
    block_size = flash.ioctl(_IOCTL_BLOCK_SIZE, 0)
    block_start = random.randrange(0, block_count)
    print(f"Raw filesystem block layout: 0 .. {block_start} .. {block_count}")

    # Copy the file to the special raw filesystem.
    _copy_file_to_raw_filesystem(filename, flash, block_start)

    # Enter mboot with a request to do a filesystem-load update.
    # Note: the filename doesn't mean anything here, but still needs to be non-empty.
    fwupdate.update_mpy(
        filename,
        fs_type=fwupdate.VFS_RAW,
        fs_base=_MBOOT_SPIFLASH_BASE_ADDR + _RAW_FILESYSTEM_ADDR + block_start * block_size,
        fs_len=(block_count - block_start) * block_size,
        fs_base2=_MBOOT_SPIFLASH_BASE_ADDR + _RAW_FILESYSTEM_ADDR,
        fs_len2=block_start * block_size,
    )



# test match.span() for unicode strings

try:
    import re
except ImportError:
    print("SKIP")
    raise SystemExit

try:
    m = re.match(".", "a")
    m.span
except AttributeError:
    print("SKIP")
    raise SystemExit


def print_spans(match):
    print("----")
    try:
        i = 0
        while True:
            print(match.span(i), match.start(i), match.end(i))
            i += 1
    except IndexError:
        pass


m = re.match(r"([0-9]*)(([a-z]*)([0-9]*))", "1234\u2764567")
print_spans(m)



try:
    from collections import namedtuple
except ImportError:
    print("SKIP")
    raise SystemExit

t = namedtuple("Tup", ["baz", "foo", "bar"])(3, 2, 5)

try:
    t._asdict
except AttributeError:
    print("SKIP")
    raise SystemExit

d = t._asdict()
print(list(d.keys()))
print(list(d.values()))



# test await expression

import sys
if sys.implementation.name == 'micropython':
    # uPy allows normal generators to be awaitables
    coroutine = lambda f: f
else:
    import types
    coroutine = types.coroutine

@coroutine
def wait(value):
    print('wait value:', value)
    msg = yield 'message from wait({})'.format(value)
    print('wait got back:', msg)
    return 10

async def f():
    x = await wait(1)**2
    print('x =', x)

coro = f()
print('return from send:', coro.send(None))
try:
    coro.send('message from main')
except StopIteration:
    print('got StopIteration')



# This example demonstrates a simple temperature sensor peripheral.
#
# The sensor's local value updates every second, and it will notify
# any connected central every 10 seconds.

# This example demonstrates the low-level bluetooth module. For most
# applications, we recommend using the higher-level aioble library which takes
# care of all IRQ handling and connection management. See
# https://github.com/micropython/micropython-lib/tree/master/micropython/bluetooth/aioble
# and in particular the temp_sensor.py example included with aioble.

import bluetooth
import random
import struct
import time
from ble_advertising import advertising_payload

from micropython import const

_IRQ_CENTRAL_CONNECT = const(1)
_IRQ_CENTRAL_DISCONNECT = const(2)
_IRQ_GATTS_INDICATE_DONE = const(20)

_FLAG_READ = const(0x0002)
_FLAG_NOTIFY = const(0x0010)
_FLAG_INDICATE = const(0x0020)

# org.bluetooth.service.environmental_sensing
_ENV_SENSE_UUID = bluetooth.UUID(0x181A)
# org.bluetooth.characteristic.temperature
_TEMP_CHAR = (
    bluetooth.UUID(0x2A6E),
    _FLAG_READ | _FLAG_NOTIFY | _FLAG_INDICATE,
)
_ENV_SENSE_SERVICE = (
    _ENV_SENSE_UUID,
    (_TEMP_CHAR,),
)

# org.bluetooth.characteristic.gap.appearance.xml
_ADV_APPEARANCE_GENERIC_THERMOMETER = const(768)


class BLETemperature:
    def __init__(self, ble, name="mpy-temp"):
        self._ble = ble
        self._ble.active(True)
        self._ble.irq(self._irq)
        ((self._handle,),) = self._ble.gatts_register_services((_ENV_SENSE_SERVICE,))
        self._connections = set()
        self._payload = advertising_payload(
            name=name, services=[_ENV_SENSE_UUID], appearance=_ADV_APPEARANCE_GENERIC_THERMOMETER
        )
        self._advertise()

    def _irq(self, event, data):
        # Track connections so we can send notifications.
        if event == _IRQ_CENTRAL_CONNECT:
            conn_handle, _, _ = data
            self._connections.add(conn_handle)
        elif event == _IRQ_CENTRAL_DISCONNECT:
            conn_handle, _, _ = data
            self._connections.remove(conn_handle)
            # Start advertising again to allow a new connection.
            self._advertise()
        elif event == _IRQ_GATTS_INDICATE_DONE:
            conn_handle, value_handle, status = data

    def set_temperature(self, temp_deg_c, notify=False, indicate=False):
        # Data is sint16 in degrees Celsius with a resolution of 0.01 degrees Celsius.
        # Write the local value, ready for a central to read.
        self._ble.gatts_write(self._handle, struct.pack("<h", int(temp_deg_c * 100)))
        if notify or indicate:
            for conn_handle in self._connections:
                if notify:
                    # Notify connected centrals.
                    self._ble.gatts_notify(conn_handle, self._handle)
                if indicate:
                    # Indicate connected centrals.
                    self._ble.gatts_indicate(conn_handle, self._handle)

    def _advertise(self, interval_us=500000):
        self._ble.gap_advertise(interval_us, adv_data=self._payload)


def demo():
    ble = bluetooth.BLE()
    temp = BLETemperature(ble)

    t = 25
    i = 0

    while True:
        # Write every second, notify every 10 seconds.
        i = (i + 1) % 10
        temp.set_temperature(t, notify=i == 0, indicate=False)
        # Random walk the temperature.
        t += random.uniform(-0.5, 0.5)
        time.sleep_ms(1000)


if __name__ == "__main__":
    demo()



# MicroPython asyncio module
# MIT license; Copyright (c) 2019-2020 Damien P. George

# This file contains the core TaskQueue based on a pairing heap, and the core Task class.
# They can optionally be replaced by C implementations.

from . import core


# pairing-heap meld of 2 heaps; O(1)
def ph_meld(h1, h2):
    if h1 is None:
        return h2
    if h2 is None:
        return h1
    lt = core.ticks_diff(h1.ph_key, h2.ph_key) < 0
    if lt:
        if h1.ph_child is None:
            h1.ph_child = h2
        else:
            h1.ph_child_last.ph_next = h2
        h1.ph_child_last = h2
        h2.ph_next = None
        h2.ph_rightmost_parent = h1
        return h1
    else:
        h1.ph_next = h2.ph_child
        h2.ph_child = h1
        if h1.ph_next is None:
            h2.ph_child_last = h1
            h1.ph_rightmost_parent = h2
        return h2


# pairing-heap pairing operation; amortised O(log N)
def ph_pairing(child):
    heap = None
    while child is not None:
        n1 = child
        child = child.ph_next
        n1.ph_next = None
        if child is not None:
            n2 = child
            child = child.ph_next
            n2.ph_next = None
            n1 = ph_meld(n1, n2)
        heap = ph_meld(heap, n1)
    return heap


# pairing-heap delete of a node; stable, amortised O(log N)
def ph_delete(heap, node):
    if node is heap:
        child = heap.ph_child
        node.ph_child = None
        return ph_pairing(child)
    # Find parent of node
    parent = node
    while parent.ph_next is not None:
        parent = parent.ph_next
    parent = parent.ph_rightmost_parent
    # Replace node with pairing of its children
    if node is parent.ph_child and node.ph_child is None:
        parent.ph_child = node.ph_next
        node.ph_next = None
        return heap
    elif node is parent.ph_child:
        child = node.ph_child
        next = node.ph_next
        node.ph_child = None
        node.ph_next = None
        node = ph_pairing(child)
        parent.ph_child = node
    else:
        n = parent.ph_child
        while node is not n.ph_next:
            n = n.ph_next
        child = node.ph_child
        next = node.ph_next
        node.ph_child = None
        node.ph_next = None
        node = ph_pairing(child)
        if node is None:
            node = n
        else:
            n.ph_next = node
    node.ph_next = next
    if next is None:
        node.ph_rightmost_parent = parent
        parent.ph_child_last = node
    return heap


# TaskQueue class based on the above pairing-heap functions.
class TaskQueue:
    def __init__(self):
        self.heap = None

    def peek(self):
        return self.heap

    def push(self, v, key=None):
        assert v.ph_child is None
        assert v.ph_next is None
        v.data = None
        v.ph_key = key if key is not None else core.ticks()
        self.heap = ph_meld(v, self.heap)

    def pop(self):
        v = self.heap
        assert v.ph_next is None
        self.heap = ph_pairing(v.ph_child)
        v.ph_child = None
        return v

    def remove(self, v):
        self.heap = ph_delete(self.heap, v)


# Task class representing a coroutine, can be waited on and cancelled.
class Task:
    def __init__(self, coro, globals=None):
        self.coro = coro  # Coroutine of this Task
        self.data = None  # General data for queue it is waiting on
        self.state = True  # None, False, True, a callable, or a TaskQueue instance
        self.ph_key = 0  # Pairing heap
        self.ph_child = None  # Paring heap
        self.ph_child_last = None  # Paring heap
        self.ph_next = None  # Paring heap
        self.ph_rightmost_parent = None  # Paring heap

    def __iter__(self):
        if not self.state:
            # Task finished, signal that is has been await'ed on.
            self.state = False
        elif self.state is True:
            # Allocated head of linked list of Tasks waiting on completion of this task.
            self.state = TaskQueue()
        elif type(self.state) is not TaskQueue:
            # Task has state used for another purpose, so can't also wait on it.
            raise RuntimeError("can't wait")
        return self

    def __next__(self):
        if not self.state:
            # Task finished, raise return value to caller so it can continue.
            raise self.data
        else:
            # Put calling task on waiting queue.
            self.state.push(core.cur_task)
            # Set calling task's data to this task that it waits on, to double-link it.
            core.cur_task.data = self

    def done(self):
        return not self.state

    def cancel(self):
        # Check if task is already finished.
        if not self.state:
            return False
        # Can't cancel self (not supported yet).
        if self is core.cur_task:
            raise RuntimeError("can't cancel self")
        # If Task waits on another task then forward the cancel to the one it's waiting on.
        while isinstance(self.data, Task):
            self = self.data
        # Reschedule Task as a cancelled task.
        if hasattr(self.data, "remove"):
            # Not on the main running queue, remove the task from the queue it's on.
            self.data.remove(self)
            core._task_queue.push(self)
        elif core.ticks_diff(self.ph_key, core.ticks()) > 0:
            # On the main running queue but scheduled in the future, so bring it forward to now.
            core._task_queue.remove(self)
            core._task_queue.push(self)
        self.data = core.CancelledError
        return True



# test that socket.connect() has correct polling behaviour before, during and after

import socket, select


def test(peer_addr):
    s = socket.socket()
    poller = select.poll()
    poller.register(s)

    # test poll before connect
    p = poller.poll(0)
    print(len(p), p[0][-1])

    s.connect(peer_addr)

    # test poll during connection
    print(len(poller.poll(0)))

    # test poll after connection is established
    p = poller.poll(1000)
    print(len(p), p[0][-1])

    s.close()


if __name__ == "__main__":
    test(socket.getaddrinfo("micropython.org", 80)[0][-1])



# Implementation of USB DFU device in Python.
#
# To run, just execute this file on a device with machine.USBDevice support.  The device
# will then change to DFU mode.
#
# For example, use `mpremote` (the `--no-follow` option starts the script running
# without waiting for a response, because there won't be a response, the USB will change
# to a DFU device):
#
# $ mpremote run --no-follow usb_dfu_device.py
#
# Then you can access the DFU device using the `pydfu.py` script in this repository, to
# list DFU device, copy a file to the device, then exit DFU mode:
#
# $ ../../tools/pydfu.py -l
# $ ../../tools/pydfu.py -u <file.dfu>
#
# After running the last command above, the USB CDC device and REPL should reappear.

import struct, machine

# USB constants for bmRequestType.
USB_REQ_RECIP_INTERFACE = 0x01
USB_REQ_TYPE_CLASS = 0x20
USB_DIR_OUT = 0x00
USB_DIR_IN = 0x80

# String describing the memory layout of the DFU device.
MEMORY_LAYOUT = b"@Internal Flash  /0x08000000/16*128Kg"

# VID and PID of the DFU device (these are the ST values).
VID = 0x0483
PID = 0xDF11

# Maximum transfer size for RX and TX.
wTransferSize = 2048

# DFU device descriptor.
_desc_dev = bytes(
    [
        0x12,  # bLength
        0x01,  # bDescriptorType: Device
        0x00,
        0x02,  # USB version: 2.00
        0x00,  # bDeviceClass
        0x00,  # bDeviceSubClass
        0x00,  # bDeviceProtocol
        0x40,  # bMaxPacketSize
        VID & 0xFF,
        VID >> 8,  # VID
        PID & 0xFF,
        PID >> 8,  # PID
        0x00,
        0x01,  # bcdDevice: 1.00
        0x11,  # iManufacturer
        0x12,  # iProduct
        0x13,  # iSerialNumber
        0x01,  # bNumConfigurations: 1
    ]
)

# DFU configuration descriptor.
_desc_cfg = bytes(
    [
        # Configuration Descriptor.
        0x09,  # bLength
        0x02,  # bDescriptorType
        0x1B,
        0x00,  # wTotalLength: 27
        0x01,  # bNumInterfaces
        0x01,  # bConfigurationValue
        0x00,  # iConfiguration
        0x80,  # bmAttributes (bus powered)
        0x32,  # bMaxPower
        # Interface Descriptor.
        0x09,  # bLength
        0x04,  # bDescriptorType
        0x00,  # bInterfaceNumber
        0x00,  # bNumEndpointns
        0x00,  # bAlternateSetting
        0xFE,  # bInterfaceClass: application specific interface
        0x01,  # bInterfaceSubClasse: device firmware update
        0x02,  # bInterfaceProtocol
        0x14,  # iInterface
        # Device Firmware Upgrade Interface Descriptor.
        0x09,  # bLength
        0x21,  # bDescriptorType
        0x0B,  # bmAttributes (will detach, upload supported, download supported)
        0xFF,
        0x00,  # wDetatchTimeout
        wTransferSize & 0xFF,
        wTransferSize >> 8,  # wTransferSize
        0x1A,
        0x01,  # bcdDFUVersion
    ]
)

# DFU strings.
_desc_strs = {
    0x11: b"iManufacturer",
    0x12: b"iProduct",
    0x13: b"iSerialNumber",
    0x14: MEMORY_LAYOUT,
}


# This class handles the DFU USB device logic.
class DFUOverUSB:
    def __init__(self, dfu):
        # USB buffer for transfers.
        self.usb_buf = bytearray(wTransferSize)
        # Instance of the DFU state machine.
        self.dfu = dfu

    def _control_xfer_cb(self, stage, request):
        bmRequestType, bRequest, wValue, wIndex, wLength = struct.unpack("<BBHHH", request)
        if stage == 1:  # SETUP
            if bmRequestType == USB_DIR_OUT | USB_REQ_TYPE_CLASS | USB_REQ_RECIP_INTERFACE:
                # Data coming from host, prepare to receive it.
                return memoryview(self.usb_buf)[:wLength]
            if bmRequestType == USB_DIR_IN | USB_REQ_TYPE_CLASS | USB_REQ_RECIP_INTERFACE:
                # Host requests data, prepare to send it.
                buf = memoryview(self.usb_buf)[:wLength]
                return self.dfu.handle_tx(bRequest, wValue, buf)
        elif stage == 3:  # ACK
            if bmRequestType & USB_DIR_IN:
                # EP0 TX sent.
                self.dfu.process()
            else:
                # EP0 RX ready.
                buf = memoryview(self.usb_buf)[:wLength]
                self.dfu.handle_rx(bRequest, wValue, buf)
        return True


# This class handles the DFU state machine.
class DFU:
    # DFU class requests.
    DETACH = 0
    DNLOAD = 1
    UPLOAD = 2
    GETSTATUS = 3
    CLRSTATUS = 4
    GETSTATE = 5
    ABORT = 6

    # DFU states.
    STATE_IDLE = 2
    STATE_BUSY = 4
    STATE_DNLOAD_IDLE = 5
    STATE_MANIFEST = 7
    STATE_UPLOAD_IDLE = 9
    STATE_ERROR = 0xA

    # DFU commands.
    CMD_NONE = 0
    CMD_EXIT = 1
    CMD_UPLOAD = 7
    CMD_DNLOAD = 8

    # Download sub-commands.
    CMD_DNLOAD_SET_ADDRESS = 0x21
    CMD_DNLOAD_ERASE = 0x41
    CMD_DNLOAD_READ_UNPROTECT = 0x92

    # Error status flags.
    STATUS_OK = 0x00

    def __init__(self):
        self.state = DFU.STATE_IDLE
        self.cmd = DFU.CMD_NONE
        self.status = DFU.STATUS_OK
        self.error = 0
        self.leave_dfu = False
        self.addr = 0
        self.dnload_block_num = 0
        self.dnload_len = 0
        self.dnload_buf = bytearray(wTransferSize)

    def handle_rx(self, cmd, arg, buf):
        # Handle an incoming packet of data.
        if cmd == DFU.CLRSTATUS:
            self.state = DFU.STATE_IDLE
            self.cmd = DFU.CMD_NONE
            self.status = DFU.STATUS_OK
            self.error = 0
        elif cmd == DFU.ABORT:
            self.state = DFU.STATE_IDLE
            self.cmd = DFU.CMD_NONE
            self.status = DFU.STATUS_OK
            self.error = 0
        elif cmd == DFU.DNLOAD:
            if len(buf) == 0:
                # Exit DFU.
                self.cmd = DFU.CMD_EXIT
            else:
                # Download data to device.
                self.cmd = DFU.CMD_DNLOAD
                self.dnload_block_num = arg
                self.dnload_len = len(buf)
                self.dnload_buf[: len(buf)] = buf

    def handle_tx(self, cmd, arg, buf):
        # Prepare data to go to the host.
        if cmd == DFU.UPLOAD:
            if arg >= 2:
                self.cmd = DFU.CMD_UPLOAD
                addr = (arg - 2) * len(buf) + self.addr
                self.do_read(addr, buf)
                return buf
            return None
        elif cmd == DFU.GETSTATUS and len(buf) == 6:
            if self.cmd == DFU.CMD_NONE:
                pass
            elif self.cmd == DFU.CMD_EXIT:
                self.state = DFU.STATE_MANIFEST
            elif self.cmd == DFU.CMD_UPLOAD:
                self.state = DFU.STATE_UPLOAD_IDLE
            elif self.cmd == DFU.CMD_DNLOAD:
                self.state = DFU.STATE_BUSY
            else:
                self.state = DFU.STATE_BUSY

            # Populate the buffer to return to the host.
            buf[0] = self.status
            buf[1] = 0
            buf[2] = 0
            buf[3] = 0
            buf[4] = self.state
            buf[5] = self.error

            # Clear errors now they've been sent to host.
            self.status = DFU.STATUS_OK
            self.error = 0

            return buf
        else:
            return None

    def process(self):
        # Transition the DFU state machine.
        if self.state == DFU.STATE_MANIFEST:
            self.leave_dfu = True
        elif self.state == DFU.STATE_BUSY:
            if self.cmd == DFU.CMD_DNLOAD:
                self.cmd = DFU.CMD_NONE
                self.state = self.process_dnload()

    def process_dnload(self):
        ret = -1  # Assume error.
        if self.dnload_block_num == 0:
            # Download control commands.
            if self.dnload_len >= 1 and self.dnload_buf[0] == DFU.CMD_DNLOAD_ERASE:
                if self.dnload_len == 1:
                    # Mass erase.
                    ret = self.do_mass_erase()
                    if ret != 0:
                        self.cmd = DFU.CMD_NONE
                elif self.dnload_len == 5:
                    # Erase page.
                    addr = struct.unpack_from("<L", self.dnload_buf, 1)[0]
                    ret = self.do_page_erase(addr)
            elif self.dnload_len >= 1 and self.dnload_buf[0] == DFU.CMD_DNLOAD_SET_ADDRESS:
                if self.dnload_len == 5:
                    # Set address.
                    self.addr = struct.unpack_from("<L", self.dnload_buf, 1)[0]
                    ret = 0
        elif self.dnload_block_num > 1:
            # Write data to memory.
            addr = (self.dnload_block_num - 2) * wTransferSize + self.addr
            ret = self.do_write(addr, self.dnload_len, self.dnload_buf)
        if ret == 0:
            return DFU.STATE_DNLOAD_IDLE
        else:
            return DFU.STATE_ERROR

    def do_mass_erase(self):
        # This function would implement a mass erase of flash memory.
        return 0  # indicate success

    def do_page_erase(self, addr):
        # This function would implement an erase of a page in flash memory.
        return 0  # indicate success

    def do_read(self, addr, buf):
        # This function would implement a read at the given address of flash memory.
        # Return some dummy bytes.
        for i in range(len(buf)):
            buf[i] = i & 0xFF
        return 0  # indicate success

    def do_write(self, addr, size, buf):
        # This function would implement a write of the given data to flash memory.
        return 0  # indicate success


# Create an instance of the DFU state machine.
dfu = DFU()

# Create an instance of the DFU USB handler.
dfu_usb = DFUOverUSB(dfu)

# Switch the USB device to the custom DFU driver.
usbd = machine.USBDevice()
usbd.active(0)
usbd.builtin_driver = usbd.BUILTIN_NONE
usbd.config(
    desc_dev=_desc_dev,
    desc_cfg=_desc_cfg,
    desc_strs=_desc_strs,
    control_xfer_cb=dfu_usb._control_xfer_cb,
)
usbd.active(1)

# Wait for the DFU state machine to complete.
while not dfu.leave_dfu:
    machine.idle()

# Switch the USB device back to the default built-in driver.
usbd.active(0)
usbd.builtin_driver = usbd.BUILTIN_DEFAULT
usbd.config(
    desc_dev=usbd.builtin_driver.desc_dev,
    desc_cfg=usbd.builtin_driver.desc_cfg,
    desc_strs=(),
)
usbd.active(1)



"""Utilities for with-statement contexts.  See PEP 343.

Original source code: https://hg.python.org/cpython/file/3.4/Lib/contextlib.py

Not implemented:
 - redirect_stdout;
 - ExitStack.
 - closing
 - supress
"""


class ContextDecorator(object):
    "A base class or mixin that enables context managers to work as decorators."

    def _recreate_cm(self):
        """Return a recreated instance of self.

        Allows an otherwise one-shot context manager like
        _GeneratorContextManager to support use as
        a decorator via implicit recreation.

        This is a private interface just for _GeneratorContextManager.
        See issue #11647 for details.
        """
        return self

    def __call__(self, func):
        def inner(*args, **kwds):
            with self._recreate_cm():
                return func(*args, **kwds)

        return inner


class _GeneratorContextManager(ContextDecorator):
    """Helper for @contextmanager decorator."""

    def __init__(self, func, *args, **kwds):
        self.gen = func(*args, **kwds)
        self.func, self.args, self.kwds = func, args, kwds

    def _recreate_cm(self):
        # _GCM instances are one-shot context managers, so the
        # CM must be recreated each time a decorated function is
        # called
        return self.__class__(self.func, *self.args, **self.kwds)

    def __enter__(self):
        try:
            return next(self.gen)
        except StopIteration:
            raise RuntimeError("generator didn't yield") from None

    def __exit__(self, type, value, traceback):
        if type is None:
            try:
                next(self.gen)
            except StopIteration:
                return
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                # Need to force instantiation so we can reliably
                # tell if we get the same exception back
                value = type()
            try:
                self.gen.throw(type, value, traceback)
                raise RuntimeError("generator didn't stop after throw()")
            except StopIteration as exc:
                # Suppress the exception *unless* it's the same exception that
                # was passed to throw().  This prevents a StopIteration
                # raised inside the "with" statement from being suppressed
                return exc is not value


def contextmanager(func):
    """@contextmanager decorator.

    Typical usage:

        @contextmanager
        def some_generator(<arguments>):
            <setup>
            try:
                yield <value>
            finally:
                <cleanup>

    This makes this:

        with some_generator(<arguments>) as <variable>:
            <body>

    equivalent to this:

        <setup>
        try:
            <variable> = <value>
            <body>
        finally:
            <cleanup>

    """

    def helper(*args, **kwds):
        return _GeneratorContextManager(func, *args, **kwds)

    return helper



#!/usr/bin/env python3
import argparse
import os
import os.path

argparser = argparse.ArgumentParser(description="Compile all .py files to .mpy recursively")
argparser.add_argument("-o", "--out", help="output directory (default: input dir)")
argparser.add_argument("--target", help="select MicroPython target config")
argparser.add_argument("dir", help="input directory")
args = argparser.parse_args()

TARGET_OPTS = {
    "unix": "",
    "baremetal": "",
}

args.dir = args.dir.rstrip("/")

if not args.out:
    args.out = args.dir

path_prefix_len = len(args.dir) + 1

for path, subdirs, files in os.walk(args.dir):
    for f in files:
        if f.endswith(".py"):
            fpath = path + "/" + f
            # print(fpath)
            out_fpath = args.out + "/" + fpath[path_prefix_len:-3] + ".mpy"
            out_dir = os.path.dirname(out_fpath)
            if not os.path.isdir(out_dir):
                os.makedirs(out_dir)
            cmd = "mpy-cross -v -v %s -s %s %s -o %s" % (
                TARGET_OPTS.get(args.target, ""),
                fpath[path_prefix_len:],
                fpath,
                out_fpath,
            )
            # print(cmd)
            res = os.system(cmd)
            assert res == 0



# MicroPython LoRa synchronous modem driver
# MIT license; Copyright (c) 2023 Angus Gratton
#
# LoRa is a registered trademark or service mark of Semtech Corporation or its affiliates.

import machine
import time


class SyncModem:
    # Mixin-like base class that provides synchronous modem send and recv
    # functions
    #
    #
    # Don't instantiate this class directly, instantiate one of the 'AsyncXYZ'
    # modem classes defined in the lora module.
    #
    # These are intended for simple applications. They block the caller until
    # the modem operation is complete, and don't support interleaving send
    # and receive.

    def _after_init(self):
        pass  # Needed for AsyncModem but not SyncModem

    def send(self, packet, tx_at_ms=None):
        # Send the given packet (byte sequence),
        # and return once transmission of the packet is complete.
        #
        # Returns a timestamp (result of time.ticks_ms()) when the packet
        # finished sending.
        self.prepare_send(packet)

        # If the caller specified a timestamp to start transmission at, wait until
        # that time before triggering the send
        if tx_at_ms is not None:
            time.sleep_ms(max(0, time.ticks_diff(tx_at_ms, time.ticks_ms())))

        will_irq = self.start_send()  # ... and go!

        # sleep for the expected send time before checking if send has ended
        time.sleep_ms(self.get_time_on_air_us(len(packet)) // 1000)

        tx = True
        while tx is True:
            self._sync_wait(will_irq)
            tx = self.poll_send()
        return tx

    def recv(self, timeout_ms=None, rx_length=0xFF, rx_packet=None):
        # Attempt to a receive a single LoRa packet, timeout after timeout_ms milliseconds
        # or wait indefinitely if no timeout is supplied (default).
        #
        # Returns an instance of RxPacket or None if the radio timed out while receiving.
        #
        # Optional rx_length argument is only used if lora_cfg["implict_header"] == True
        # (not the default) and holds the length of the payload to receive.
        #
        # Optional rx_packet argument can be an existing instance of RxPacket
        # which will be reused to save allocations, but only if the received packet
        # is the same length as the rx_packet packet. If the length is different, a
        # new RxPacket instance is allocated and returned.
        will_irq = self.start_recv(timeout_ms, False, rx_length)
        rx = True
        while rx is True:
            self._sync_wait(will_irq)
            rx = self.poll_recv(rx_packet)
        return rx or None

    def _sync_wait(self, will_irq):
        # For synchronous usage, block until an interrupt occurs or we time out
        if will_irq:
            for n in range(100):
                machine.idle()
                # machine.idle() wakes up very often, so don't actually return
                # unless _radio_isr ran already. The outer for loop is so the
                # modem is still polled occasionally to
                # avoid the possibility an IRQ was lost somewhere.
                #
                # None of this is very efficient, power users should either use
                # async or call the low-level API manually with better
                # port-specific sleep configurations, in order to get the best
                # efficiency.
                if self.irq_triggered():
                    break
        else:
            time.sleep_ms(1)



from functools import reduce

res = reduce(lambda x, y: x + y, [1, 2, 3, 4, 5])
assert res == 1 + 2 + 3 + 4 + 5

res = reduce(lambda x, y: x + y, [1, 2, 3, 4, 5], 10)
assert res == 10 + 1 + 2 + 3 + 4 + 5



# NeoPixel driver for MicroPython
# MIT license; Copyright (c) 2016 Damien P. George, 2021 Jim Mussared

from machine import bitstream


class NeoPixel:
    # G R B W
    ORDER = (1, 0, 2, 3)

    def __init__(self, pin, n, bpp=3, timing=1):
        self.pin = pin
        self.n = n
        self.bpp = bpp
        self.buf = bytearray(n * bpp)
        self.pin.init(pin.OUT)
        # Timing arg can either be 1 for 800kHz or 0 for 400kHz,
        # or a user-specified timing ns tuple (high_0, low_0, high_1, low_1).
        self.timing = (
            ((400, 850, 800, 450) if timing else (800, 1700, 1600, 900))
            if isinstance(timing, int)
            else timing
        )

    def __len__(self):
        return self.n

    def __setitem__(self, i, v):
        offset = i * self.bpp
        for i in range(self.bpp):
            self.buf[offset + self.ORDER[i]] = v[i]

    def __getitem__(self, i):
        offset = i * self.bpp
        return tuple(self.buf[offset + self.ORDER[i]] for i in range(self.bpp))

    def fill(self, v):
        b = self.buf
        l = len(self.buf)
        bpp = self.bpp
        for i in range(bpp):
            c = v[i]
            j = self.ORDER[i]
            while j < l:
                b[j] = c
                j += bpp

    def write(self):
        # BITSTREAM_TYPE_HIGH_LOW = 0
        bitstream(self.pin, 0, self.timing, self.buf)



# check that consts are not replaced in anything except standalone identifiers

from micropython import const

X = const(1)
Y = const(2)
Z = const(3)

# import that uses a constant
import micropython as X

print(globals()["X"])


# function name that matches a constant
def X():
    print("function X", X)


globals()["X"]()


# arguments that match a constant
def f(X, *Y, **Z):
    pass


f(1)


# class name that matches a constant
class X:
    def f(self):
        print("class X", X)


globals()["X"]().f()


# constant within a class
class A:
    C1 = const(4)

    def X(self):
        print("method X", Y, C1, self.C1)


A().X()



import time
from multiprocessing import Pool


def f(x):
    return x * x


pool = Pool(4)
future = pool.apply_async(f, (10,))
assert future.get() == 100


def f2(x):
    time.sleep(0.5)
    return x + 1


future = pool.apply_async(f2, (10,))
iter = 0
while not future.ready():
    # print("not ready")
    time.sleep(0.1)
    iter += 1

assert future.get() == 11
assert iter >= 5 and iter <= 8

t = time.time()
futs = [
    pool.apply_async(f2, (10,)),
    pool.apply_async(f2, (11,)),
    pool.apply_async(f2, (12,)),
]

iter = 0
while True:
    # not all(futs):
    c = 0
    for f in futs:
        if not f.ready():
            c += 1
    if not c:
        break
    # print("not ready2")
    time.sleep(0.1)
    iter += 1

assert iter >= 5 and iter <= 8

print("Run 3 parallel sleep(1)'s in: ", time.time() - t)



import unittest
from ucontextlib import contextmanager


class ContextManagerTestCase(unittest.TestCase):
    def setUp(self):
        self._history = []

        @contextmanager
        def manager(x):
            self._history.append("start")
            try:
                yield x
            finally:
                self._history.append("finish")

        self._manager = manager

    def test_context_manager(self):
        with self._manager(123) as x:
            self.assertEqual(x, 123)
        self.assertEqual(self._history, ["start", "finish"])

    def test_context_manager_on_error(self):
        exc = Exception()
        try:
            with self._manager(123):
                raise exc
        except Exception as e:
            self.assertEqual(exc, e)
        self.assertEqual(self._history, ["start", "finish"])


if __name__ == "__main__":
    unittest.main()



# Test for VfsPosix error conditions

try:
    import os, sys, vfs

    vfs.VfsPosix
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit

if sys.platform == "win32":
    # Windows doesn't let you delete the current directory, so this cannot be
    # tested.
    print("SKIP")
    raise SystemExit

# We need an empty directory for testing.
# Skip the test if it already exists.
temp_dir = "vfs_posix_enoent_test_dir"
try:
    os.stat(temp_dir)
    print("SKIP")
    raise SystemExit
except OSError:
    pass

curdir = os.getcwd()
os.mkdir(temp_dir)
os.chdir(temp_dir)
os.rmdir(curdir + "/" + temp_dir)
try:
    print("getcwd():", os.getcwd())
except OSError as e:
    # expecting ENOENT = 2
    print("getcwd():", repr(e))

try:
    print("VfsPosix():", vfs.VfsPosix("something"))
except OSError as e:
    # expecting ENOENT = 2
    print("VfsPosix():", repr(e))



# test various type conversions

import micropython


# converting incoming arg to bool
@micropython.viper
def f1(x: bool):
    print(x)


f1(0)
f1(1)
f1([])
f1([1])


# taking and returning a bool
@micropython.viper
def f2(x: bool) -> bool:
    return x


print(f2([]))
print(f2([1]))


# converting to bool within function
@micropython.viper
def f3(x) -> bool:
    return bool(x)


print(f3([]))
print(f3(-1))



import tls
from tls import *


class SSLContext:
    def __init__(self, *args):
        self._context = tls.SSLContext(*args)
        self._context.verify_mode = CERT_NONE

    @property
    def verify_mode(self):
        return self._context.verify_mode

    @verify_mode.setter
    def verify_mode(self, val):
        self._context.verify_mode = val

    def load_cert_chain(self, certfile, keyfile):
        if isinstance(certfile, str):
            with open(certfile, "rb") as f:
                certfile = f.read()
        if isinstance(keyfile, str):
            with open(keyfile, "rb") as f:
                keyfile = f.read()
        self._context.load_cert_chain(certfile, keyfile)

    def load_verify_locations(self, cafile=None, cadata=None):
        if cafile:
            with open(cafile, "rb") as f:
                cadata = f.read()
        self._context.load_verify_locations(cadata)

    def wrap_socket(
        self, sock, server_side=False, do_handshake_on_connect=True, server_hostname=None
    ):
        return self._context.wrap_socket(
            sock,
            server_side=server_side,
            do_handshake_on_connect=do_handshake_on_connect,
            server_hostname=server_hostname,
        )


def wrap_socket(
    sock,
    server_side=False,
    key=None,
    cert=None,
    cert_reqs=CERT_NONE,
    cadata=None,
    server_hostname=None,
    do_handshake=True,
):
    con = SSLContext(PROTOCOL_TLS_SERVER if server_side else PROTOCOL_TLS_CLIENT)
    if cert or key:
        con.load_cert_chain(cert, key)
    if cadata:
        con.load_verify_locations(cadata=cadata)
    con.verify_mode = cert_reqs
    return con.wrap_socket(
        sock,
        server_side=server_side,
        do_handshake_on_connect=do_handshake,
        server_hostname=server_hostname,
    )



from esp32 import Partition

# MicroPython's partition table uses "vfs", TinyUF2 uses "ffat".
bdev = Partition.find(Partition.TYPE_DATA, label="vfs")
if not bdev:
    bdev = Partition.find(Partition.TYPE_DATA, label="ffat", block_size=512)
bdev = bdev[0] if bdev else None



# test that we can generate a traceback without allocating

import micropython
import sys

try:
    import io
except ImportError:
    print("SKIP")
    raise SystemExit

# preallocate exception instance with some room for a traceback
global_exc = StopIteration()
try:
    raise global_exc
except:
    pass


def test():
    micropython.heap_lock()
    global global_exc
    global_exc.__traceback__ = None
    try:
        raise global_exc
    except StopIteration:
        print("StopIteration")
    micropython.heap_unlock()


# call test() with heap allocation disabled
test()

# print the exception that was raised
buf = io.StringIO()
sys.print_exception(global_exc, buf)
for l in buf.getvalue().split("\n"):
    # uPy on pyboard prints <stdin> as file, so remove filename.
    if l.startswith("  File "):
        l = l.split('"')
        print(l[0], l[2])
    else:
        print(l)



# test using lock to coordinate access to global mutable objects
#
# MIT license; Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd

import time
import _thread


def fac(n):
    x = 1
    for i in range(1, n + 1):
        x *= i
    return x


def thread_entry():
    while True:
        with jobs_lock:
            try:
                f, arg = jobs.pop(0)
            except IndexError:
                return
        ans = f(arg)
        with output_lock:
            output.append((arg, ans))


# create a list of jobs
jobs = [(fac, i) for i in range(20, 80)]
jobs_lock = _thread.allocate_lock()
n_jobs = len(jobs)

# create a list to store the results
output = []
output_lock = _thread.allocate_lock()

# spawn threads to do the jobs
for i in range(4):
    try:
        _thread.start_new_thread(thread_entry, ())
    except OSError:
        # System cannot create a new thead, so stop trying to create them.
        break

# wait for the jobs to complete
while True:
    with jobs_lock:
        if len(output) == n_jobs:
            break
    time.sleep(0)

# sort and print the results
output.sort(key=lambda x: x[0])
for arg, ans in output:
    print(arg, ans)



f = open("data/file1")

with f as f2:
    print(f2.read())

# File should be closed
try:
    f.read()
except:
    # Note: CPython and us throw different exception trying to read from
    # close file.
    print("can't read file after with")


# Regression test: test that exception in with initialization properly
# thrown and doesn't crash.
try:
    with open("__non_existent", "r"):
        pass
except OSError:
    print("OSError")



import io
a = io.BytesIO(b"foobar")
try:
    a.seek(-10)
except Exception as e:
    # CPython throws ValueError, but MicroPython has consistent stream
    # interface, so BytesIO raises the same error as a real file, which
    # is OSError(EINVAL).
    print(type(e), e.args[0] > 0)



# test if conditions which are optimised by the compiler

if 0:
    print(5)
else:
    print(6)

if 1:
    print(7)

if 2:
    print(8)

if -1:
    print(9)
elif 1:
    print(10)

if 0:
    print(11)
else:
    print(12)

if 0:
    print(13)
elif 1:
    print(14)

if 0:
    print(15)
elif 0:
    print(16)
else:
    print(17)

if not False:
    print('a')

if not True:
    print('a')
else:
    print('b')

if False:
    print('a')
else:
    print('b')

if True:
    print('a')

if (1,):
    print('a')

if not (1,):
    print('a')
else:
    print('b')

# test evaluation of the if-condition with tuples as arguments
# non-constant tuples should be evaluated even though they will evaluate to true

def f(x):
    print("f", x)

if (f(1),):
    print(18)

if (f(2), f(3)):
    print(19)

# test if-conditions within a function

f2 = 0

def f(t1, t2, f1):
    if False:
        print(1)
    if True:
        print(1)
    if ():
        print(1)
    if (1,):
        print(1)
    if (1, 2):
        print(1)
    if t1 and t2:
        print(1)
    if (t1 and t2): # parsed differently to above
        print(1)
    if not (t1 and f1):
        print(1)
    if t1 or t2:
        print(1)
    if (t1 or t2): # parse differently to above
        print(1)
    if f1 or t1:
        print(1)
    if not (f1 or f2):
        print(1)
    if t1 and f1 or t1 and t2:
        print(1)
    if (f1 or t1) and (f2 or t2):
        print(1)

f(True, 1, False)



import ffi
import struct
import os
import errno
import ffilib
import utime
import math
from uselect import *


libc = ffilib.libc()

# int epoll_create(int size);
epoll_create = libc.func("i", "epoll_create", "i")
# int epoll_ctl(int epfd, int op, int fd, struct epoll_event *event);
epoll_ctl = libc.func("i", "epoll_ctl", "iiiP")
# int epoll_wait(int epfd, struct epoll_event *events, int maxevents, int timeout);
epoll_wait = libc.func("i", "epoll_wait", "ipii")

EPOLLIN = 0x001
EPOLLPRI = 0x002
EPOLLOUT = 0x004
EPOLLERR = 0x008
EPOLLHUP = 0x010
EPOLLRDHUP = 0x2000
EPOLLONESHOT = 1 << 30
EPOLLET = 1 << 31

EPOLL_CTL_ADD = 1
EPOLL_CTL_DEL = 2
EPOLL_CTL_MOD = 3

# Not included in uselect.
POLLPRI = 0x002

# TODO: struct epoll_event's 2nd member is union of uint64_t, etc.
# On x86, uint64_t is 4-byte aligned, on many other platforms - 8-byte.
# Until uctypes module can assign native struct offset, use dirty hack
# below.
# TODO: Get rid of all this dirtiness, move it on C side
if ffilib.bitness > 32:
    # On x86_64, epoll_event is packed struct
    epoll_event = "<IO"
elif struct.calcsize("IQ") == 12:
    epoll_event = "IO"
else:
    epoll_event = "QO"


class Epoll:
    def __init__(self, epfd):
        self.epfd = epfd
        self.evbuf = struct.pack(epoll_event, 0, None)
        self.registry = {}

    def register(self, fd, eventmask=EPOLLIN | EPOLLPRI | EPOLLOUT, retval=None):
        "retval is extension to stdlib, value to use in results from .poll()."
        if retval is None:
            retval = fd
        s = struct.pack(epoll_event, eventmask, retval)
        r = epoll_ctl(self.epfd, EPOLL_CTL_ADD, fd, s)
        if r == -1 and os.errno_() == errno.EEXIST:
            r = epoll_ctl(self.epfd, EPOLL_CTL_MOD, fd, s)
        os.check_error(r)
        # We must keep reference to retval, or it may be GCed. And we must
        # keep mapping from fd to retval to be able to get rid of this retval
        # reference later.
        self.registry[fd] = retval

    def unregister(self, fd):
        # Pass dummy event structure, to workaround kernel bug
        r = epoll_ctl(self.epfd, EPOLL_CTL_DEL, fd, self.evbuf)
        os.check_error(r)
        del self.registry[fd]

    def poll_ms(self, timeout=-1):
        s = bytearray(self.evbuf)
        if timeout >= 0:
            deadline = utime.ticks_add(utime.ticks_ms(), timeout)
        while True:
            n = epoll_wait(self.epfd, s, 1, timeout)
            if not os.check_error(n):
                break
            if timeout >= 0:
                timeout = utime.ticks_diff(deadline, utime.ticks_ms())
                if timeout < 0:
                    n = 0
                    break
        res = []
        if n > 0:
            vals = struct.unpack(epoll_event, s)
            res.append((vals[1], vals[0]))
        return res

    def poll(self, timeout=-1):
        return self.poll_ms(-1 if timeout == -1 else math.ceil(timeout * 1000))

    def close(self):
        os.close(self.epfd)


def epoll(sizehint=4):
    fd = epoll_create(sizehint)
    os.check_error(fd)
    return Epoll(fd)



# Test for shutting down and restarting the BLE stack.

import sys

# ruff: noqa: E402
sys.path.append("")

from micropython import const
import machine
import time

import asyncio
import aioble
import bluetooth

TIMEOUT_MS = 5000

SERVICE_UUID = bluetooth.UUID("A5A5A5A5-FFFF-9999-1111-5A5A5A5A5A5A")
CHAR_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444444")

_L2CAP_PSN = const(22)
_L2CAP_MTU = const(128)


# Acting in peripheral role.
async def instance0_task():
    multitest.globals(BDADDR=aioble.config("mac"))
    multitest.next()

    for i in range(3):
        service = aioble.Service(SERVICE_UUID)
        characteristic = aioble.Characteristic(service, CHAR_UUID, read=True)
        aioble.register_services(service)

        # Write initial characteristic value.
        characteristic.write("periph{}".format(i))

        multitest.broadcast("connect-{}".format(i))

        # Wait for central to connect to us.
        print("advertise")
        connection = await aioble.advertise(
            20_000, adv_data=b"\x02\x01\x06\x04\xffMPY", timeout_ms=TIMEOUT_MS
        )
        print("connected")

        multitest.broadcast("connected-{}".format(i))

        for j in range(3):
            channel = await connection.l2cap_accept(_L2CAP_PSN, _L2CAP_MTU)
            print("channel accepted")

            buf = bytearray(10)
            n = await channel.recvinto(buf)
            print("recv", n, buf[:n])

            multitest.broadcast("recv-{}-{}".format(i, j))

            await channel.disconnected(5000)
            print("channel disconnected")

        # Wait for the central to disconnect.
        await connection.disconnected(timeout_ms=TIMEOUT_MS)
        print("disconnected")

        # Shutdown aioble + modbluetooth.
        print("shutdown")
        aioble.stop()


def instance0():
    try:
        asyncio.run(instance0_task())
    finally:
        aioble.stop()


# Acting in central role.
async def instance1_task():
    multitest.next()

    for i in range(3):
        multitest.wait("connect-{}".format(i))
        # Connect to peripheral.
        print("connect")
        device = aioble.Device(*BDADDR)
        connection = await device.connect(timeout_ms=TIMEOUT_MS)

        multitest.wait("connected-{}".format(i))

        # Discover characteristics.
        service = await connection.service(SERVICE_UUID)
        print("service", service.uuid)
        characteristic = await service.characteristic(CHAR_UUID)
        print("characteristic", characteristic.uuid)

        # Issue read of characteristic, should get initial value.
        print("read", await characteristic.read(timeout_ms=TIMEOUT_MS))

        for j in range(3):
            print("connecting channel")
            channel = await connection.l2cap_connect(_L2CAP_PSN, _L2CAP_MTU)
            print("channel connected")

            await channel.send("l2cap-{}-{}".format(i, j))
            await channel.flush()

            multitest.wait("recv-{}-{}".format(i, j))

            print("disconnecting channel")
            await channel.disconnect()
            print("channel disconnected")

            await asyncio.sleep_ms(100)

        # Disconnect from peripheral.
        print("disconnect")
        await connection.disconnect(timeout_ms=TIMEOUT_MS)
        print("disconnected")

        # Shutdown aioble.
        print("shutdown")
        aioble.stop()


def instance1():
    try:
        asyncio.run(instance1_task())
    finally:
        aioble.stop()



from machine import Pin

p = Pin("SW1", Pin.IN)
if p.value() == 1:
    print("pass")
else:
    print("FAIL")

p = Pin("LED1", Pin.OUT)
p.high()
if p.value() == 1:
    print("pass")
else:
    print("FAIL")
p.low()



import ffilib


SIG_DFL = 0
SIG_IGN = 1

SIGINT = 2
SIGPIPE = 13
SIGTERM = 15

libc = ffilib.libc()

signal_i = libc.func("i", "signal", "ii")
signal_p = libc.func("i", "signal", "ip")


def signal(n, handler):
    if isinstance(handler, int):
        return signal_i(n, handler)
    import ffi

    cb = ffi.callback("v", handler, "i")
    return signal_p(n, cb)



# Test GATTC/S data transfer between peripheral and central, and use of gatts_set_buffer()

from micropython import const
import time, machine, bluetooth

TIMEOUT_MS = 5000

_IRQ_CENTRAL_CONNECT = const(1)
_IRQ_CENTRAL_DISCONNECT = const(2)
_IRQ_GATTS_WRITE = const(3)
_IRQ_PERIPHERAL_CONNECT = const(7)
_IRQ_PERIPHERAL_DISCONNECT = const(8)
_IRQ_GATTC_SERVICE_RESULT = const(9)
_IRQ_GATTC_SERVICE_DONE = const(10)
_IRQ_GATTC_CHARACTERISTIC_RESULT = const(11)
_IRQ_GATTC_CHARACTERISTIC_DONE = const(12)
_IRQ_GATTC_READ_RESULT = const(15)
_IRQ_GATTC_READ_DONE = const(16)
_IRQ_GATTC_WRITE_DONE = const(17)
_IRQ_GATTC_NOTIFY = const(18)

SERVICE_UUID = bluetooth.UUID("00000001-1111-2222-3333-444444444444")
CHAR_CTRL_UUID = bluetooth.UUID("00000002-1111-2222-3333-444444444444")
CHAR_RX_UUID = bluetooth.UUID("00000003-1111-2222-3333-444444444444")
CHAR_TX_UUID = bluetooth.UUID("00000004-1111-2222-3333-444444444444")
CHAR_CTRL = (CHAR_CTRL_UUID, bluetooth.FLAG_WRITE | bluetooth.FLAG_NOTIFY)
CHAR_RX = (CHAR_RX_UUID, bluetooth.FLAG_WRITE | bluetooth.FLAG_WRITE_NO_RESPONSE)
CHAR_TX = (CHAR_TX_UUID, bluetooth.FLAG_NOTIFY)
SERVICE = (SERVICE_UUID, (CHAR_CTRL, CHAR_RX, CHAR_TX))
SERVICES = (SERVICE,)

waiting_events = {}


def irq(event, data):
    if event == _IRQ_CENTRAL_CONNECT:
        print("_IRQ_CENTRAL_CONNECT")
        waiting_events[event] = data[0]
    elif event == _IRQ_CENTRAL_DISCONNECT:
        print("_IRQ_CENTRAL_DISCONNECT")
    elif event == _IRQ_GATTS_WRITE:
        print("_IRQ_GATTS_WRITE")
        waiting_events[(event, data[1])] = None
    elif event == _IRQ_PERIPHERAL_CONNECT:
        print("_IRQ_PERIPHERAL_CONNECT")
        waiting_events[event] = data[0]
    elif event == _IRQ_PERIPHERAL_DISCONNECT:
        print("_IRQ_PERIPHERAL_DISCONNECT")
    elif event == _IRQ_GATTC_CHARACTERISTIC_RESULT:
        if data[-1] == CHAR_CTRL_UUID:
            print("_IRQ_GATTC_CHARACTERISTIC_RESULT", data[-1])
            waiting_events[(event, CHAR_CTRL_UUID)] = data[2]
        elif data[-1] == CHAR_RX_UUID:
            print("_IRQ_GATTC_CHARACTERISTIC_RESULT", data[-1])
            waiting_events[(event, CHAR_RX_UUID)] = data[2]
        elif data[-1] == CHAR_TX_UUID:
            print("_IRQ_GATTC_CHARACTERISTIC_RESULT", data[-1])
            waiting_events[(event, CHAR_TX_UUID)] = data[2]
    elif event == _IRQ_GATTC_CHARACTERISTIC_DONE:
        print("_IRQ_GATTC_CHARACTERISTIC_DONE")
    elif event == _IRQ_GATTC_READ_RESULT:
        print("_IRQ_GATTC_READ_RESULT", data[-1])
    elif event == _IRQ_GATTC_READ_DONE:
        print("_IRQ_GATTC_READ_DONE", data[-1])
    elif event == _IRQ_GATTC_WRITE_DONE:
        print("_IRQ_GATTC_WRITE_DONE", data[-1])
    elif event == _IRQ_GATTC_NOTIFY:
        print("_IRQ_GATTC_NOTIFY", bytes(data[-1]))
        waiting_events[(event, data[1])] = None

    if event not in waiting_events:
        waiting_events[event] = None


def wait_for_event(event, timeout_ms):
    t0 = time.ticks_ms()
    while time.ticks_diff(time.ticks_ms(), t0) < timeout_ms:
        if event in waiting_events:
            return waiting_events.pop(event)
        machine.idle()
    raise ValueError("Timeout waiting for {}".format(event))


# Acting in peripheral role.
def instance0():
    multitest.globals(BDADDR=ble.config("mac"))
    ((char_ctrl_handle, char_rx_handle, char_tx_handle),) = ble.gatts_register_services(SERVICES)

    # Increase the size of the rx buffer and enable append mode.
    ble.gatts_set_buffer(char_rx_handle, 100, True)

    print("gap_advertise")
    ble.gap_advertise(20_000, b"\x02\x01\x06\x04\xffMPY")
    multitest.next()
    try:
        # Wait for central to connect to us.
        conn_handle = wait_for_event(_IRQ_CENTRAL_CONNECT, TIMEOUT_MS)

        # Wait for the central to signal that it's done with its part of the test.
        wait_for_event((_IRQ_GATTS_WRITE, char_ctrl_handle), 2 * TIMEOUT_MS)

        # Read all accumulated data from the central.
        print("gatts_read:", ble.gatts_read(char_rx_handle))

        # Notify the central a few times.
        for i in range(4):
            time.sleep_ms(300)
            ble.gatts_notify(conn_handle, char_tx_handle, "message{}".format(i))

        # Notify the central that we are done with our part of the test.
        time.sleep_ms(300)
        ble.gatts_notify(conn_handle, char_ctrl_handle, "OK")

        # Wait for the central to disconnect.
        wait_for_event(_IRQ_CENTRAL_DISCONNECT, TIMEOUT_MS)
    finally:
        ble.active(0)


# Acting in central role.
def instance1():
    multitest.next()
    try:
        # Connect to peripheral and then disconnect.
        print("gap_connect")
        ble.gap_connect(*BDADDR)
        conn_handle = wait_for_event(_IRQ_PERIPHERAL_CONNECT, TIMEOUT_MS)

        # Discover characteristics.
        ble.gattc_discover_characteristics(conn_handle, 1, 65535)
        wait_for_event(_IRQ_GATTC_CHARACTERISTIC_DONE, TIMEOUT_MS)
        ctrl_value_handle = waiting_events[(_IRQ_GATTC_CHARACTERISTIC_RESULT, CHAR_CTRL_UUID)]
        rx_value_handle = waiting_events[(_IRQ_GATTC_CHARACTERISTIC_RESULT, CHAR_RX_UUID)]
        tx_value_handle = waiting_events[(_IRQ_GATTC_CHARACTERISTIC_RESULT, CHAR_TX_UUID)]

        # Write to the characteristic a few times, with and without response.
        for i in range(4):
            print("gattc_write")
            ble.gattc_write(conn_handle, rx_value_handle, "central{}".format(i), i & 1)
            if i & 1:
                wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
            time.sleep_ms(400)

        # Write to say that we are done with our part of the test.
        ble.gattc_write(conn_handle, ctrl_value_handle, "OK", 1)
        wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)

        # Wait for notification that peripheral is done with its part of the test.
        wait_for_event((_IRQ_GATTC_NOTIFY, ctrl_value_handle), TIMEOUT_MS)

        # Disconnect from peripheral.
        print("gap_disconnect:", ble.gap_disconnect(conn_handle))
        wait_for_event(_IRQ_PERIPHERAL_DISCONNECT, TIMEOUT_MS)
    finally:
        ble.active(0)


ble = bluetooth.BLE()
ble.active(1)
ble.irq(irq)



def gen():
    print("sent:", (yield 1))
    yield 2

def gen2():
    print((yield from gen()))

g = gen2()
next(g)
print("yielded:", g.send("val"))
try:
    next(g)
except StopIteration:
    print("StopIteration")



class C:
    def f():
        pass

# del a class attribute

del C.f
try:
    print(C.x)
except AttributeError:
    print("AttributeError")
try:
    del C.f
except AttributeError:
    print("AttributeError")

# del an instance attribute

c = C()

c.x = 1
print(c.x)

del c.x
try:
    print(c.x)
except AttributeError:
    print("AttributeError")
try:
    del c.x
except AttributeError:
    print("AttributeError")

# try to del an attribute of a built-in class
try:
    del int.to_bytes
except (AttributeError, TypeError):
    # uPy raises AttributeError, CPython raises TypeError
    print('AttributeError/TypeError')



# Driver for Mboot, the MicroPython boot loader
# MIT license; Copyright (c) 2018 Damien P. George

import struct, time, os, hashlib


I2C_CMD_ECHO = 1
I2C_CMD_GETID = 2
I2C_CMD_GETCAPS = 3
I2C_CMD_RESET = 4
I2C_CMD_CONFIG = 5
I2C_CMD_GETLAYOUT = 6
I2C_CMD_MASSERASE = 7
I2C_CMD_PAGEERASE = 8
I2C_CMD_SETRDADDR = 9
I2C_CMD_SETWRADDR = 10
I2C_CMD_READ = 11
I2C_CMD_WRITE = 12
I2C_CMD_COPY = 13
I2C_CMD_CALCHASH = 14
I2C_CMD_MARKVALID = 15


class Bootloader:
    def __init__(self, i2c, addr):
        self.i2c = i2c
        self.addr = addr
        self.buf1 = bytearray(1)
        try:
            self.i2c.writeto(addr, b"")
        except OSError:
            raise Exception("no I2C mboot device found")

    def wait_response(self):
        start = time.ticks_ms()
        while 1:
            try:
                self.i2c.readfrom_into(self.addr, self.buf1)
                n = self.buf1[0]
                break
            except OSError:
                time.sleep_us(500)
            if time.ticks_diff(time.ticks_ms(), start) > 5000:
                raise Exception("timeout")
        if n >= 129:
            raise Exception(n)
        if n == 0:
            return b""
        else:
            return self.i2c.readfrom(self.addr, n)

    def wait_empty_response(self):
        ret = self.wait_response()
        if ret:
            raise Exception("expected empty response got %r" % ret)
        else:
            return None

    def echo(self, data):
        self.i2c.writeto(self.addr, struct.pack("<B", I2C_CMD_ECHO) + data)
        return self.wait_response()

    def getid(self):
        self.i2c.writeto(self.addr, struct.pack("<B", I2C_CMD_GETID))
        ret = self.wait_response()
        unique_id = ret[:12]
        mcu_name, board_name = ret[12:].split(b"\x00")
        return unique_id, str(mcu_name, "ascii"), str(board_name, "ascii")

    def reset(self):
        self.i2c.writeto(self.addr, struct.pack("<B", I2C_CMD_RESET))
        # we don't expect any response

    def getlayout(self):
        self.i2c.writeto(self.addr, struct.pack("<B", I2C_CMD_GETLAYOUT))
        layout = self.wait_response()
        id, flash_addr, layout = layout.split(b"/")
        assert id == b"@Internal Flash  "
        flash_addr = int(flash_addr, 16)
        pages = []
        for chunk in layout.split(b","):
            n, sz = chunk.split(b"*")
            n = int(n)
            assert sz.endswith(b"Kg")
            sz = int(sz[:-2]) * 1024
            for i in range(n):
                pages.append((flash_addr, sz))
                flash_addr += sz
        return pages

    def pageerase(self, addr):
        self.i2c.writeto(self.addr, struct.pack("<BI", I2C_CMD_PAGEERASE, addr))
        self.wait_empty_response()

    def setrdaddr(self, addr):
        self.i2c.writeto(self.addr, struct.pack("<BI", I2C_CMD_SETRDADDR, addr))
        self.wait_empty_response()

    def setwraddr(self, addr):
        self.i2c.writeto(self.addr, struct.pack("<BI", I2C_CMD_SETWRADDR, addr))
        self.wait_empty_response()

    def read(self, n):
        self.i2c.writeto(self.addr, struct.pack("<BB", I2C_CMD_READ, n))
        return self.wait_response()

    def write(self, buf):
        self.i2c.writeto(self.addr, struct.pack("<B", I2C_CMD_WRITE) + buf)
        self.wait_empty_response()

    def calchash(self, n):
        self.i2c.writeto(self.addr, struct.pack("<BI", I2C_CMD_CALCHASH, n))
        return self.wait_response()

    def markvalid(self):
        self.i2c.writeto(self.addr, struct.pack("<B", I2C_CMD_MARKVALID))
        self.wait_empty_response()

    def deployfile(self, filename, addr):
        pages = self.getlayout()
        page_erased = [False] * len(pages)
        buf = bytearray(128)  # maximum payload supported by I2C protocol
        start_addr = addr
        self.setwraddr(addr)
        fsize = os.stat(filename)[6]
        local_sha = hashlib.sha256()
        print("Deploying %s to location 0x%08x" % (filename, addr))
        with open(filename, "rb") as f:
            t0 = time.ticks_ms()
            while True:
                n = f.readinto(buf)
                if n == 0:
                    break

                # check if we need to erase the page
                for i, p in enumerate(pages):
                    if p[0] <= addr < p[0] + p[1]:
                        # found page
                        if not page_erased[i]:
                            print(
                                "\r% 3u%% erase 0x%08x"
                                % (100 * (addr - start_addr) // fsize, addr),
                                end="",
                            )
                            self.pageerase(addr)
                            page_erased[i] = True
                        break
                else:
                    raise Exception("address 0x%08x not valid" % addr)

                # write the data
                self.write(buf)

                # update local SHA256, with validity bits set
                if addr == start_addr:
                    buf[0] |= 3
                if n == len(buf):
                    local_sha.update(buf)
                else:
                    local_sha.update(buf[:n])

                addr += n
                ntotal = addr - start_addr
                if ntotal % 2048 == 0 or ntotal == fsize:
                    print("\r% 3u%% % 7u bytes   " % (100 * ntotal // fsize, ntotal), end="")
            t1 = time.ticks_ms()
        print()
        print("rate: %.2f KiB/sec" % (1024 * ntotal / (t1 - t0) / 1000))

        local_sha = local_sha.digest()
        print("Local SHA256: ", "".join("%02x" % x for x in local_sha))

        self.setrdaddr(start_addr)
        remote_sha = self.calchash(ntotal)
        print("Remote SHA256:", "".join("%02x" % x for x in remote_sha))

        if local_sha == remote_sha:
            print("Marking app firmware as valid")
            self.markvalid()

        self.reset()



try:
    import errno, os, vfs

    vfs.VfsFat
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class RAMFS:
    SEC_SIZE = 512

    def __init__(self, blocks):
        self.data = bytearray(blocks * self.SEC_SIZE)

    def readblocks(self, n, buf):
        # print("readblocks(%s, %x(%d))" % (n, id(buf), len(buf)))
        for i in range(len(buf)):
            buf[i] = self.data[n * self.SEC_SIZE + i]

    def writeblocks(self, n, buf):
        # print("writeblocks(%s, %x)" % (n, id(buf)))
        for i in range(len(buf)):
            self.data[n * self.SEC_SIZE + i] = buf[i]

    def ioctl(self, op, arg):
        # print("ioctl(%d, %r)" % (op, arg))
        if op == 4:  # MP_BLOCKDEV_IOCTL_BLOCK_COUNT
            return len(self.data) // self.SEC_SIZE
        if op == 5:  # MP_BLOCKDEV_IOCTL_BLOCK_SIZE
            return self.SEC_SIZE


try:
    bdev = RAMFS(50)
    vfs.VfsFat.mkfs(bdev)
except MemoryError:
    print("SKIP")
    raise SystemExit

print(b"FOO_FILETXT" not in bdev.data)
print(b"hello!" not in bdev.data)

fs = vfs.VfsFat(bdev)
vfs.mount(fs, "/ramdisk")

print("statvfs:", fs.statvfs("/ramdisk"))
print("getcwd:", fs.getcwd())

try:
    fs.stat("no_file.txt")
except OSError as e:
    print(e.errno == errno.ENOENT)

with fs.open("foo_file.txt", "w") as f:
    f.write("hello!")
print(list(fs.ilistdir()))

print("stat root:", fs.stat("/")[:-3])  # timestamps differ across runs
print("stat file:", fs.stat("foo_file.txt")[:-3])  # timestamps differ across runs

print(b"FOO_FILETXT" in bdev.data)
print(b"hello!" in bdev.data)

fs.mkdir("foo_dir")
fs.chdir("foo_dir")
print("getcwd:", fs.getcwd())
print(list(fs.ilistdir()))

with fs.open("sub_file.txt", "w") as f:
    f.write("subdir file")

try:
    fs.chdir("sub_file.txt")
except OSError as e:
    print(e.errno == errno.ENOENT)

fs.chdir("..")
print("getcwd:", fs.getcwd())

vfs.umount(fs)

fs = vfs.VfsFat(bdev)
print(list(fs.ilistdir(b"")))

# list a non-existent directory
try:
    fs.ilistdir(b"no_exist")
except OSError as e:
    print("ENOENT:", e.errno == errno.ENOENT)



import logging

# Create logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Create console handler and set level to debug
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.DEBUG)

# Create file handler and set level to error
file_handler = logging.FileHandler("error.log", mode="w")
file_handler.setLevel(logging.ERROR)

# Create a formatter
formatter = logging.Formatter("%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s")

# Add formatter to the handlers
stream_handler.setFormatter(formatter)
file_handler.setFormatter(formatter)

# Add handlers to logger
logger.addHandler(stream_handler)
logger.addHandler(file_handler)

# Log some messages
logger.debug("debug message")
logger.info("info message")
logger.warning("warn message")
logger.error("error message")
logger.critical("critical message")
logger.info("message %s %d", "arg", 5)
logger.info("message %(foo)s %(bar)s", {"foo": 1, "bar": 20})

try:
    1 / 0
except:
    logger.error("Some trouble (%s)", "expected")


# Custom handler example
class MyHandler(logging.Handler):
    def emit(self, record):
        print("levelname=%(levelname)s name=%(name)s message=%(message)s" % record.__dict__)


logging.getLogger().addHandler(MyHandler())
logging.info("Test message7")



"""
This pre-processor parses a single file containing a list of
`MP_REGISTER_MODULE(MP_QSTR_module_name, obj_module)` or
`MP_REGISTER_EXTENSIBLE_MODULE(MP_QSTR_module_name, obj_module)`
(i.e. the output of `py/makeqstrdefs.py cat module`).

The output is a header (typically moduledefs.h) which is included by
py/objmodule.c that contains entries to be included in the definition of
 - mp_rom_map_elem_t mp_builtin_module_table[]
 - mp_rom_map_elem_t mp_builtin_extensible_module_table[]

Extensible modules are modules that can be overridden from the filesystem, see
py/builtinimnport.c:process_import_at_level. Regular modules will always use
the built-in version.
"""

from __future__ import print_function

import sys
import re
import io
import argparse


register_pattern = re.compile(
    r"\s*(MP_REGISTER_MODULE|MP_REGISTER_EXTENSIBLE_MODULE)\(MP_QSTR_(.*?),\s*(.*?)\);",
    flags=re.DOTALL,
)

delegation_pattern = re.compile(
    r"\s*(?:MP_REGISTER_MODULE_DELEGATION)\((.*?),\s*(.*?)\);",
    flags=re.DOTALL,
)


def find_module_registrations(filename):
    """Find any MP_REGISTER_MODULE definitions in the provided file.

    :param str filename: path to file to check
    :return: List[(module_name, obj_module)]
    """
    global pattern

    with io.open(filename, encoding="utf-8") as c_file_obj:
        c = c_file_obj.read()
        return set(re.findall(register_pattern, c)), set(re.findall(delegation_pattern, c))


def generate_module_table_header(modules):
    """Generate header with module table entries for builtin modules.

    :param List[(module_name, obj_module)] modules: module defs
    :return: None
    """

    # Print header file for all external modules.
    mod_defs = set()
    extensible_mod_defs = set()
    for macro_name, module_name, obj_module in modules:
        mod_def = "MODULE_DEF_{}".format(module_name.upper())
        if macro_name == "MP_REGISTER_MODULE":
            mod_defs.add(mod_def)
        elif macro_name == "MP_REGISTER_EXTENSIBLE_MODULE":
            extensible_mod_defs.add(mod_def)
        if "," in obj_module:
            print(
                "ERROR: Call to {}({}, {}) should be {}({}, {})\n".format(
                    macro_name,
                    module_name,
                    obj_module,
                    macro_name,
                    module_name,
                    obj_module.split(",")[0],
                ),
                file=sys.stderr,
            )
            sys.exit(1)
        print(
            (
                "extern const struct _mp_obj_module_t {obj_module};\n"
                "#undef {mod_def}\n"
                "#define {mod_def} {{ MP_ROM_QSTR(MP_QSTR_{module_name}), MP_ROM_PTR(&{obj_module}) }},\n"
            ).format(
                module_name=module_name,
                obj_module=obj_module,
                mod_def=mod_def,
            )
        )

    print("\n#define MICROPY_REGISTERED_MODULES \\")

    for mod_def in sorted(mod_defs):
        print("    {mod_def} \\".format(mod_def=mod_def))

    print("// MICROPY_REGISTERED_MODULES")

    print("\n#define MICROPY_REGISTERED_EXTENSIBLE_MODULES \\")

    for mod_def in sorted(extensible_mod_defs):
        print("    {mod_def} \\".format(mod_def=mod_def))

    print("// MICROPY_REGISTERED_EXTENSIBLE_MODULES")


def generate_module_delegations(delegations):
    if not delegations:
        return

    print()
    for obj_module, fun_name in delegations:
        print("extern void {}(mp_obj_t self_in, qstr attr, mp_obj_t *dest);".format(fun_name))
    print("#define MICROPY_MODULE_DELEGATIONS \\")
    for obj_module, fun_name in delegations:
        print(
            "    {{ MP_ROM_PTR(&{obj_module}), {fun_name} }}, \\".format(
                obj_module=obj_module, fun_name=fun_name
            )
        )
    print("// MICROPY_MODULE_DELEGATIONS")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("file", nargs=1, help="file with MP_REGISTER_MODULE definitions")
    args = parser.parse_args()

    print("// Automatically generated by makemoduledefs.py.\n")

    modules, delegations = find_module_registrations(args.file[0])
    generate_module_table_header(sorted(modules))
    generate_module_delegations(sorted(delegations))


if __name__ == "__main__":
    main()



# test that viper functions capture their globals context

gl = {}

exec(
    """
@micropython.viper
def f():
    return x
""",
    gl,
)

# x is not yet in the globals, f should not see it
try:
    print(gl["f"]())
except NameError:
    print("NameError")

# x is in globals, f should now see it
gl["x"] = 123
print(gl["f"]())



import os


dir = "."
if "/" in __file__:
    dir = __file__.rsplit("/", 1)[0]

assert os.access(dir + "/test_filestat.py", os.F_OK) == True
assert os.access(dir + "/test_filestat.py-not", os.F_OK) == False



"""
categories: Types,Exception
description: Exception.__init__ method does not exist.
cause: Subclassing native classes is not fully supported in MicroPython.
workaround: Call using ``super()`` instead::

    class A(Exception):
        def __init__(self):
            super().__init__()
"""


class A(Exception):
    def __init__(self):
        Exception.__init__(self)


a = A()



# test bignum operation with float/complex

i = 1 << 65

# convert bignum to float on rhs
print("%.5g" % (2.0 * i))

# negative bignum as float
print("%.5g" % float(-i))

# this should convert to float
print("%.5g" % (i / 5))

# these should delegate to float
print("%.5g" % (i * 1.2))
print("%.5g" % (i / 1.2))

# this should delegate to complex
print("%.5g" % (i * 1.2j).imag)

# negative power should produce float
print("%.5g" % (i**-1))
print("%.5g" % ((2 + i - i) ** -3))

try:
    i / 0
except ZeroDivisionError:
    print("ZeroDivisionError")



# Test asyncio.gather() function, features that are not implemented.

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit


def custom_handler(loop, context):
    print(repr(context["exception"]))


async def task(id):
    print("task start", id)
    await asyncio.sleep(0.01)
    print("task end", id)
    return id


async def gather_task(t0, t1):
    print("gather_task start")
    await asyncio.gather(t0, t1)
    print("gather_task end")


async def main():
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(custom_handler)

    # Test case where can't wait on a task being gathered.
    tasks = [asyncio.create_task(task(1)), asyncio.create_task(task(2))]
    gt = asyncio.create_task(gather_task(tasks[0], tasks[1]))
    await asyncio.sleep(0)  # let the gather start
    try:
        await tasks[0]  # can't await because this task is part of the gather
    except RuntimeError as er:
        print(repr(er))
    await gt

    print("====")

    # Test case where can't gather on a task being waited.
    tasks = [asyncio.create_task(task(1)), asyncio.create_task(task(2))]
    asyncio.create_task(gather_task(tasks[0], tasks[1]))
    await tasks[0]  # wait on this task before the gather starts
    await tasks[1]


asyncio.run(main())



# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Glenn Ruben Bakke
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
MicroPython Seeedstudio TFT Shield V2 driver, SPI interfaces, Analog GPIO
Contains SD-card reader, LCD and Touch sensor

The pca10040 pin layout is used as reference.

Example usage of LCD:

    from seeedstudio_tft_shield_v2 import ILI9341

    lcd = ILI9341(240, 320)
    lcd.text("Hello World!, 32, 32)
    lcd.show()

Example usage of SD card reader:

    import os
    from seeedstudio_tft_shield_v2 import mount_tf

    tf = mount_tf()
    os.listdir()
"""
import vfs
import time
import framebuf

from machine import SPI, Pin
from sdcard import SDCard


def mount_tf(self, mount_point="/"):
    sd = SDCard(SPI(0), Pin("P15", mode=Pin.OUT))
    vfs.mount(sd, mount_point)


class ILI9341:
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.pages = self.height // 8
        self.buffer = bytearray(self.pages * self.width)
        self.framebuf = framebuf.FrameBuffer(
            self.buffer, self.width, self.height, framebuf.MONO_VLSB
        )

        self.spi = SPI(0)
        # chip select
        self.cs = Pin("P16", mode=Pin.OUT, pull=Pin.PULL_UP)
        # command
        self.dc = Pin("P17", mode=Pin.OUT, pull=Pin.PULL_UP)

        # initialize all pins high
        self.cs.high()
        self.dc.high()

        self.spi.init(baudrate=8000000, phase=0, polarity=0)

        self.init_display()

    def init_display(self):
        time.sleep_ms(500)

        self.write_cmd(0x01)

        time.sleep_ms(200)

        self.write_cmd(0xCF)
        self.write_data(bytearray([0x00, 0x8B, 0x30]))

        self.write_cmd(0xED)
        self.write_data(bytearray([0x67, 0x03, 0x12, 0x81]))

        self.write_cmd(0xE8)
        self.write_data(bytearray([0x85, 0x10, 0x7A]))

        self.write_cmd(0xCB)
        self.write_data(bytearray([0x39, 0x2C, 0x00, 0x34, 0x02]))

        self.write_cmd(0xF7)
        self.write_data(bytearray([0x20]))

        self.write_cmd(0xEA)
        self.write_data(bytearray([0x00, 0x00]))

        # Power control
        self.write_cmd(0xC0)
        # VRH[5:0]
        self.write_data(bytearray([0x1B]))

        # Power control
        self.write_cmd(0xC1)
        # SAP[2:0];BT[3:0]
        self.write_data(bytearray([0x10]))

        # VCM control
        self.write_cmd(0xC5)
        self.write_data(bytearray([0x3F, 0x3C]))

        # VCM control2
        self.write_cmd(0xC7)
        self.write_data(bytearray([0xB7]))

        # Memory Access Control
        self.write_cmd(0x36)
        self.write_data(bytearray([0x08]))

        self.write_cmd(0x3A)
        self.write_data(bytearray([0x55]))

        self.write_cmd(0xB1)
        self.write_data(bytearray([0x00, 0x1B]))

        # Display Function Control
        self.write_cmd(0xB6)
        self.write_data(bytearray([0x0A, 0xA2]))

        # 3Gamma Function Disable
        self.write_cmd(0xF2)
        self.write_data(bytearray([0x00]))

        # Gamma curve selected
        self.write_cmd(0x26)
        self.write_data(bytearray([0x01]))

        # Set Gamma
        self.write_cmd(0xE0)
        self.write_data(
            bytearray(
                [
                    0x0F,
                    0x2A,
                    0x28,
                    0x08,
                    0x0E,
                    0x08,
                    0x54,
                    0xA9,
                    0x43,
                    0x0A,
                    0x0F,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                ]
            )
        )

        # Set Gamma
        self.write_cmd(0xE1)
        self.write_data(
            bytearray(
                [
                    0x00,
                    0x15,
                    0x17,
                    0x07,
                    0x11,
                    0x06,
                    0x2B,
                    0x56,
                    0x3C,
                    0x05,
                    0x10,
                    0x0F,
                    0x3F,
                    0x3F,
                    0x0F,
                ]
            )
        )

        # Exit Sleep
        self.write_cmd(0x11)
        time.sleep_ms(120)

        # Display on
        self.write_cmd(0x29)
        time.sleep_ms(500)
        self.fill(0)

    def show(self):
        # set col
        self.write_cmd(0x2A)
        self.write_data(bytearray([0x00, 0x00]))
        self.write_data(bytearray([0x00, 0xEF]))

        # set page
        self.write_cmd(0x2B)
        self.write_data(bytearray([0x00, 0x00]))
        self.write_data(bytearray([0x01, 0x3F]))

        self.write_cmd(0x2C)

        for row in range(0, self.pages):
            for pixel_pos in range(0, 8):
                for col in range(0, self.width):
                    compressed_pixel = self.buffer[row * 240 + col]
                    if ((compressed_pixel >> pixel_pos) & 0x1) == 0:
                        self.write_data(bytearray([0x00, 0x00]))
                    else:
                        self.write_data(bytearray([0xFF, 0xFF]))

    def fill(self, col):
        self.framebuf.fill(col)

    def pixel(self, x, y, col):
        self.framebuf.pixel(x, y, col)

    def scroll(self, dx, dy):
        self.framebuf.scroll(dx, dy)

    def text(self, string, x, y, col=1):
        self.framebuf.text(string, x, y, col)

    def write_cmd(self, cmd):
        self.dc.low()
        self.cs.low()
        self.spi.write(bytearray([cmd]))
        self.cs.high()

    def write_data(self, buf):
        self.dc.high()
        self.cs.low()
        self.spi.write(buf)
        self.cs.high()



# Test for VfsPosix with relative paths

try:
    import os, vfs

    vfs.VfsPosix
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit

# We need a directory for testing that doesn't already exist.
# Skip the test if it does exist.
temp_dir = "vfs_posix_paths_test_dir"
try:
    os.stat(temp_dir)
    print("SKIP")
    raise SystemExit
except OSError:
    pass

curdir = os.getcwd()
os.mkdir(temp_dir)

# construct new VfsPosix with absolute path argument
temp_dir_abs = os.getcwd() + os.sep + temp_dir
fs = vfs.VfsPosix(temp_dir_abs)
# when VfsPosix is used the intended way via vfs.mount(), it can only be called
# with relative paths when the CWD is inside or at its root, so simulate that
os.chdir(temp_dir_abs)
fs.mkdir("subdir")
fs.mkdir("subdir/one")
print('listdir("/"):', sorted(i[0] for i in fs.ilistdir("/")))
print('listdir("."):', sorted(i[0] for i in fs.ilistdir(".")))
print('getcwd() in {"", "/"}:', fs.getcwd() in {"", "/"})
print('chdir("subdir"):', fs.chdir("subdir"))
print("getcwd():", fs.getcwd())
print('mkdir("two"):', fs.mkdir("two"))
f = fs.open("file.py", "w")
f.write("print('hello')")
f.close()
print('listdir("/"):', sorted(i[0] for i in fs.ilistdir("/")))
print('listdir("/subdir"):', sorted(i[0] for i in fs.ilistdir("/subdir")))
print('listdir("."):', sorted(i[0] for i in fs.ilistdir(".")))
try:
    f = fs.open("/subdir/file.py", "r")
    print(f.read())
    f.close()
except Exception as e:
    print(e)
import sys

sys.path.insert(0, "")
try:
    import file

    print(file)
except Exception as e:
    print(e)
del sys.path[0]
fs.remove("file.py")
fs.rmdir("two")
fs.rmdir("/subdir/one")
fs.chdir("/")
fs.rmdir("/subdir")

# done with fs, restore CWD
os.chdir(curdir)

# some integration tests with a mounted VFS
vfs.mount(vfs.VfsPosix(temp_dir_abs), "/mnt")
os.mkdir("/mnt/dir")
print('chdir("/mnt/dir"):', os.chdir("/mnt/dir"))
print("getcwd():", os.getcwd())
print('chdir("/mnt"):', os.chdir("/mnt"))
print("getcwd():", os.getcwd())
print('chdir("/"):', os.chdir("/"))
print("getcwd():", os.getcwd())
print('chdir("/mnt/dir"):', os.chdir("/mnt/dir"))
print("getcwd():", os.getcwd())
print('chdir(".."):', os.chdir(".."))
print("getcwd():", os.getcwd())
os.rmdir("/mnt/dir")
vfs.umount("/mnt")

# restore CWD
os.chdir(curdir)

# rmdir
os.rmdir(temp_dir)
print(temp_dir in os.listdir())



# Combine bootloader, partition table and application into a final binary.

import os, sys

sys.path.append(os.getenv("IDF_PATH") + "/components/partition_table")

import gen_esp32part

OFFSET_BOOTLOADER_DEFAULT = 0x1000
OFFSET_PARTITIONS_DEFAULT = 0x8000


def load_sdkconfig_value(filename, value, default):
    value = "CONFIG_" + value + "="
    with open(filename, "r") as f:
        for line in f:
            if line.startswith(value):
                return line.split("=", 1)[1]
    return default


def load_sdkconfig_hex_value(filename, value, default):
    value = load_sdkconfig_value(filename, value, None)
    if value is None:
        return default
    return int(value, 16)


def load_sdkconfig_str_value(filename, value, default):
    value = load_sdkconfig_value(filename, value, None)
    if value is None:
        return default
    return value.strip().strip('"')


def load_partition_table(filename):
    with open(filename, "rb") as f:
        return gen_esp32part.PartitionTable.from_binary(f.read())


# Extract command-line arguments.
arg_sdkconfig = sys.argv[1]
arg_bootloader_bin = sys.argv[2]
arg_partitions_bin = sys.argv[3]
arg_application_bin = sys.argv[4]
arg_output_bin = sys.argv[5]
arg_output_uf2 = sys.argv[6]

# Load required sdkconfig values.
idf_target = load_sdkconfig_str_value(arg_sdkconfig, "IDF_TARGET", "").upper()
offset_bootloader = load_sdkconfig_hex_value(
    arg_sdkconfig, "BOOTLOADER_OFFSET_IN_FLASH", OFFSET_BOOTLOADER_DEFAULT
)
offset_partitions = load_sdkconfig_hex_value(
    arg_sdkconfig, "PARTITION_TABLE_OFFSET", OFFSET_PARTITIONS_DEFAULT
)

# Load the partition table.
partition_table = load_partition_table(arg_partitions_bin)

max_size_bootloader = offset_partitions - offset_bootloader
max_size_partitions = 0
offset_application = 0
max_size_application = 0

# Inspect the partition table to find offsets and maximum sizes.
for part in partition_table:
    if part.name == "nvs":
        max_size_partitions = part.offset - offset_partitions
    elif part.type == gen_esp32part.APP_TYPE and offset_application == 0:
        offset_application = part.offset
        max_size_application = part.size

# Define the input files, their location and maximum size.
files_in = [
    ("bootloader", offset_bootloader, max_size_bootloader, arg_bootloader_bin),
    ("partitions", offset_partitions, max_size_partitions, arg_partitions_bin),
    ("application", offset_application, max_size_application, arg_application_bin),
]
file_out = arg_output_bin

# Write output file with combined firmware.
cur_offset = offset_bootloader
with open(file_out, "wb") as fout:
    for name, offset, max_size, file_in in files_in:
        assert offset >= cur_offset
        fout.write(b"\xff" * (offset - cur_offset))
        cur_offset = offset
        with open(file_in, "rb") as fin:
            data = fin.read()
            fout.write(data)
            cur_offset += len(data)
            print(
                "%-12s@0x%06x % 8d  (% 8d remaining)"
                % (name, offset, len(data), max_size - len(data))
            )
            if len(data) > max_size:
                print(
                    "ERROR: %s overflows allocated space of %d bytes by %d bytes"
                    % (name, max_size, len(data) - max_size)
                )
                sys.exit(1)
    print("%-22s% 8d" % ("total", cur_offset))

# Generate .uf2 file if the SoC has native USB.
if idf_target in ("ESP32S2", "ESP32S3"):
    sys.path.append(os.path.join(os.path.dirname(__file__), "../../tools"))
    import uf2conv

    families = uf2conv.load_families()
    uf2conv.appstartaddr = 0
    uf2conv.familyid = families[idf_target]
    with open(arg_application_bin, "rb") as fin, open(arg_output_uf2, "wb") as fout:
        fout.write(uf2conv.convert_to_uf2(fin.read()))



try:
    import array
except ImportError:
    print("SKIP")
    raise SystemExit

print(array.array('b', [1, 2]) in b'\x01\x02\x03')
# CPython gives False here
#print(b"\x01\x02\x03" == array.array("B", [1, 2, 3]))



try:
    import heapq
except ImportError:
    print("SKIP")
    raise SystemExit

try:
    heapq.heappop([])
except IndexError:
    print("IndexError")

try:
    heapq.heappush((), 1)
except TypeError:
    print("TypeError")


def pop_and_print(h):
    l = []
    while h:
        l.append(str(heapq.heappop(h)))
    print(" ".join(l))


h = []
heapq.heappush(h, 3)
heapq.heappush(h, 1)
heapq.heappush(h, 2)
print(h)
pop_and_print(h)

h = [4, 3, 8, 9, 10, 2, 7, 11, 5]
heapq.heapify(h)
print(h)
heapq.heappush(h, 1)
heapq.heappush(h, 6)
heapq.heappush(h, 12)
print(h)
pop_and_print(h)



import gettext

msg = gettext.gettext("yes")
assert msg == "yes"

msg = gettext.ngettext("one", "two", 1)
assert msg == "one"

msg = gettext.ngettext("one", "two", 2)
assert msg == "two"

msg = gettext.ngettext("one", "two", 0)
assert msg == "two"

msg = gettext.ngettext("one", "two", "three")
assert msg == "two"



b1 = b"long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes long bytes"
b2 = b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes" b"concatenated bytes"
